# Copyright (c) The PyAMF Project.
# See LICENSE.txt for details.

"""
Tests for L{miniamf.version}
"""

import unittest

from miniamf import versions


class VersionTestCase(unittest.TestCase):
    """
    Tests for L{miniamf.version.get_version}
    """

    def test_version(self):
        self.assertEqual(versions.get_version((0, 0)), u"0.0")
        self.assertEqual(versions.get_version((0, 1)), u"0.1")
        self.assertEqual(versions.get_version((3, 2)), u"3.2")
        self.assertEqual(versions.get_version((3, 2, 1)), u"3.2.1")

        self.assertEqual(
            versions.get_version((3, 2, 1, b'alpha')),
            u"3.2.1alpha"
        )

        self.assertEqual(
            versions.get_version((3, 2, 1, u"final")),
            u"3.2.1final"
        )

    def test_class(self):
        V = versions.Version

        v1 = V(0, 1)

        self.assertEqual(v1, (0, 1))
        self.assertEqual(str(v1), u"0.1")

        v2 = V(3, 2, 1, u"final")

        self.assertEqual(v2, (3, 2, 1, u"final"))
        self.assertEqual(str(v2), u"3.2.1final")

        self.assertTrue(v2 > v1)



from miniamf._accel cimport codec, util, amf3


cdef class Context(codec.Context):
    cdef amf3.Context amf3_context


cdef class Decoder(codec.Decoder):
    cdef public bint use_amf3
    cdef readonly Context context
    cdef amf3.Decoder amf3_decoder

    cdef object readAMF3(self)
    cdef object readLongString(self, bint bytes=?)
    cdef object readMixedArray(self)
    cdef object readReference(self)
    cdef object readTypedObject(self)
    cdef void readObjectAttributes(self, object obj_attrs)
    cdef object readBytes(self)
    cdef object readBoolean(self)


cdef class Encoder(codec.Encoder):
    cdef public bint use_amf3
    cdef readonly Context context
    cdef amf3.Encoder amf3_encoder

    cdef inline int _writeEndObject(self) except -1
    cdef int writeAMF3(self, o) except -1
    cdef int _writeDict(self, dict attrs) except -1
    cdef inline int writeReference(self, o) except -2



# Copyright (c) The PyAMF Project.
# See LICENSE.txt for details.

"""
Unit tests.

@since: 0.1.0
"""

import unittest


if not hasattr(unittest.TestCase, 'assertIdentical'):
    def assertIdentical(self, first, second, msg=None):
        """
        Fail the test if C{first} is not C{second}.  This is an
        obect-identity-equality test, not an object equality (i.e. C{__eq__})
        test.

        @param msg: if msg is None, then the failure message will be
            '%r is not %r' % (first, second)
        """
        if first is not second:
            raise AssertionError(msg or '%r is not %r' % (first, second))

        return first

    unittest.TestCase.assertIdentical = assertIdentical

if not hasattr(unittest.TestCase, 'assertNotIdentical'):
    def assertNotIdentical(self, first, second, msg=None):
        """
        Fail the test if C{first} is C{second}.  This is an
        object-identity-equality test, not an object equality
        (i.e. C{__eq__}) test.

        @param msg: if msg is None, then the failure message will be
            '%r is %r' % (first, second)
        """
        if first is second:
            raise AssertionError(msg or '%r is %r' % (first, second))

        return first

    unittest.TestCase.assertNotIdentical = assertNotIdentical

if not hasattr(unittest.TestCase, 'patch'):
    import inspect

    def unpatch(self):
        for orig, part, replaced in self.__patches:
            setattr(orig, part, replaced)

    def patch(self, orig, replace):
        if not hasattr(self, '__patches'):
            self.__patches = []
            self.addCleanup(unpatch, self)

        f = inspect.stack()[1][0]

        parts = orig.split('.')

        v = f.f_globals.copy()
        v.update(f.f_locals)

        orig = v[parts[0]]

        for part in parts[1:-1]:
            orig = getattr(orig, part)

        to_replace = getattr(orig, parts[-1])

        self.__patches.append((orig, parts[-1], to_replace))

        setattr(orig, parts[-1], replace)

    unittest.TestCase.patch = patch



# Copyright (c) The PyAMF Project.
# See LICENSE.txt for details.

"""
Because there is disparity between Python packaging (and it is being sorted
out ...) we currently provide our own way to get the string of a version tuple.

@since: 0.6
"""


class Version(tuple):

    _version = None

    def __new__(cls, *args):
        x = tuple.__new__(cls, args)

        return x

    def __str__(self):
        if not self._version:
            self._version = get_version(self)

        return self._version


def get_version(elements):
    v = []
    first = True

    for x in elements:
        if not first and isinstance(x, int):
            v.append(".")
        if isinstance(x, str):
            v.append(x)
        elif isinstance(x, bytes):
            v.append(x.decode('utf-8'))
        else:
            v.append(str(x))
        first = False

    return "".join(v)



# -*- coding: utf-8 -*-
#
# Copyright (c) The PyAMF Project.
# See LICENSE.txt for details.

"""
Provides the pure Python version of L{BufferedByteStream}.

Do not reference directly, use L{miniamf.util.BufferedByteStream} instead.

@since: 0.6
"""

import struct
import io


def _get_endian_system():
    encoded = struct.pack("@I", 0x01020304)
    if encoded == b'\x01\x02\x03\x04':
        return ENDIAN_BIG
    elif encoded == b'\x04\x03\x02\x01':
        return ENDIAN_LITTLE
    else:
        raise ValueError("unrecognized system endianness: %r" % (encoded,))


#: Network byte order
ENDIAN_NETWORK = "!"

#: Native byte order
ENDIAN_NATIVE = "@"

#: Little endian
ENDIAN_LITTLE = "<"

#: Big endian
ENDIAN_BIG = ">"

#: System endian (whichever of "<" or ">" corresponds to the behavior of "@").
ENDIAN_SYSTEM = _get_endian_system()

#: All valid endianness strings
VALID_ENDIANS = (ENDIAN_NETWORK, ENDIAN_NATIVE, ENDIAN_LITTLE, ENDIAN_BIG)


def _compile_packers(endian):
    """
    Compile struct packers for all of the formats used by BufferedByteStream.
    Called whenever a BufferedByteStream's endianness is changed.
    """
    return {
        "B": struct.Struct(endian + "B"),
        "b": struct.Struct(endian + "b"),
        "h": struct.Struct(endian + "h"),
        "H": struct.Struct(endian + "H"),
        "l": struct.Struct(endian + "l"),
        "L": struct.Struct(endian + "L"),
        "d": struct.Struct(endian + "d"),
        "f": struct.Struct(endian + "f"),
    }


class Excursion(object):
    """
    Context manager which saves and restores the seek position of a filelike.
    """
    def __init__(self, fp):
        if not hasattr(fp, "seek") or not hasattr(fp, "tell"):
            raise TypeError("%r: not a seekable filelike" % fp)

        self._fp = fp
        self._pos = None

    def __enter__(self):
        self._pos = self._fp.tell()

    def __exit__(self, *unused):
        if self._pos is not None:
            self._fp.seek(self._pos, 0)


# Note: BytesIO in Python 2 is not a type, so it cannot be subclassed.
class BufferedByteStream(object):
    """
    I am a C{BytesIO} type object containing byte data from the AMF stream.

    Features:
     - Always read-write.
     - Raises L{IOError} if reading past end.
     - Allows you to C{peek()} into the stream.
     - Knows its length.

    @see: U{ByteArray on OSFlash
        <http://osflash.org/documentation/amf3#x0c_-_bytearray>}
    @see: U{Parsing ByteArrays on OSFlash
        <http://osflash.org/documentation/amf3/parsing_byte_arrays>}

    @ivar endian: Byte ordering used to represent the data. Default byte order
        is L{ENDIAN_NETWORK}.
    @type endian: ENDIAN_* code
    """

    def __init__(self, data=None, endian=ENDIAN_NETWORK):
        """
        @raise TypeError: C{data} is not acceptable to C{append}.
        """

        self.endian = endian
        self._buf = io.BytesIO()
        self._len = 0
        self.append(data)

    def __len__(self):
        if self._len is None:
            with Excursion(self._buf):
                self._buf.seek(0, 2)
                self._len = self._buf.tell()

        return self._len

    def seek(self, offset, whence=0):
        return self._buf.seek(offset, whence)

    def tell(self):
        return self._buf.tell()

    def truncate(self, size=0):
        self._buf.truncate(size)
        self._len = size

    def getvalue(self):
        return self._buf.getvalue()

    def read(self, length=-1):
        """
        If C{length} is -1 or unspecified, return the rest of the buffer.
        Otherwise, return exactly the specified number of bytes from the
        buffer.  Either way, advance the seek position over the data read.

        @raise IOError: Attempted to read past the end of the buffer.
        """
        if length == 0:
            return b''
        if length < -1:
            raise IOError("invalid read length: %r" % length)
        if self.at_eof():
            raise IOError(
                "Attempted to read from the buffer but already at the end")

        if length == -1:
            return self._buf.read()
        else:
            if self._buf.tell() + length > len(self):
                raise IOError(
                    "Attempted to read %d bytes from the buffer but only %d "
                    "remain" % (length, len(self) - self.tell())
                )
            return self._buf.read(length)

    def peek(self, size=1):
        """
        Looks up to C{size} bytes ahead in the stream without changing the
        seek position.  Unlike C{read}, it is not an error to try to peek
        past the end of the buffer, and fewer than the number of requested
        bytes may be returned.

        @param size: Default is 1.
        @type size: C{int}
        @raise ValueError: Trying to peek backwards.

        @return: Bytes.
        """
        if size < -1:
            raise ValueError("Cannot peek backwards")

        if size == 0:
            return b''

        if size == -1:
            size = self.remaining()

        peeked = b''
        with Excursion(self._buf):
            while len(peeked) < size:
                c = self._buf.read(1)
                if not c:
                    break
                peeked += c

        return peeked

    def consume(self):
        """
        Discard all of the data already read (from byte 0 up to C{tell()})
        and reset the seek position to the new beginning of the stream.

        @since: 0.4
        """
        rest = self._buf.read()
        self._buf.seek(0, 0)
        self._buf.truncate(0)
        self._len = 0
        self.append(rest)

    def remaining(self):
        """
        Returns number of remaining bytes.

        @rtype: C{number}
        @return: Number of remaining bytes.
        """
        return len(self) - self.tell()

    def at_eof(self):
        """
        Returns C{True} if the seek position is at the end of the stream.

        @rtype: C{bool}
        """
        return self.tell() == len(self)

    def write(self, s):
        """
        Writes the content of the specified C{s} into this buffer at the
        current seek position, and advance the seek position.

        @param s: Raw bytes
        """
        self._buf.write(s)
        self._len = None

    def append(self, data):
        """
        Append data to the end of the stream.  Does not change the
        seek position.

        @param data: The data to append to the stream.
        @type data: None, Unicode string, byte string, byte buffer, or
        any object with either a getvalue method or read+seek+tell
        methods.  In the latter two cases, the value returned by
        getvalue / read must be a Unicode string, byte string, or byte buffer.
        @raise TypeError: data is not convertible to a byte sequence.

        """

        if data is None:
            return

        odata = data
        if not isinstance(data, (bytes, str, bytearray)):
            if hasattr(data, "getvalue"):
                data = data.getvalue()
            elif (hasattr(data, "read") and
                  hasattr(data, "seek") and
                  hasattr(data, "tell")):
                with Excursion(data):
                    data = data.read()

        if isinstance(data, str):
            data = data.encode("utf-8")
        elif isinstance(data, bytearray):
            data = bytes(data)

        if not isinstance(data, bytes):
            raise TypeError("expected a string or filelike, not %r"
                            % odata)

        with Excursion(self._buf):
            self._buf.seek(0, 2)
            self._buf.write(data)
            self._len = self._buf.tell()

    def __add__(self, other):
        new = BufferedByteStream(self)
        new.append(other)
        return new

    # Methods for reading and writing typed data.
    @property
    def endian(self):
        """The endianness of this stream."""
        return self._endian

    @endian.setter
    def endian(self, val):
        if val not in VALID_ENDIANS:
            raise ValueError("invalid endianness code %r" % (val,))
        self._endian = val
        self._packers = _compile_packers(val)

    def _is_little_endian(self):
        if self._endian == ENDIAN_NATIVE:
            return ENDIAN_SYSTEM == ENDIAN_LITTLE
        return self._endian == ENDIAN_LITTLE

    def read_uchar(self):
        """
        Reads an C{unsigned char} from the stream.
        """
        return self._packers["B"].unpack(self.read(1))[0]

    def write_uchar(self, c):
        """
        Writes an C{unsigned char} to the stream.

        @param c: Unsigned char
        @type c: C{int}
        @raise TypeError: Unexpected type for int C{c}.
        @raise OverflowError: Not in range.
        """
        if not isinstance(c, int):
            raise TypeError("expected an int, got %r" % type(c))

        if not 0 <= c <= 255:
            raise OverflowError("Not in range, %d" % c)

        self.write(self._packers["B"].pack(c))

    def read_char(self):
        """
        Reads a C{char} from the stream.
        """
        return self._packers["b"].unpack(self.read(1))[0]

    def write_char(self, c):
        """
        Write a C{char} to the stream.

        @param c: char
        @type c: C{int}
        @raise TypeError: Unexpected type for int C{c}.
        @raise OverflowError: Not in range.
        """
        if not isinstance(c, int):
            raise TypeError("expected an int, got %r" % type(c))

        if not -128 <= c <= 127:
            raise OverflowError("Not in range, %d" % c)

        self.write(self._packers["b"].pack(c))

    def read_ushort(self):
        """
        Reads a 2 byte unsigned integer from the stream.
        """
        return self._packers["H"].unpack(self.read(2))[0]

    def write_ushort(self, s):
        """
        Writes a 2 byte unsigned integer to the stream.

        @param s: 2 byte unsigned integer
        @type s: C{int}
        @raise TypeError: Unexpected type for int C{s}.
        @raise OverflowError: Not in range.
        """
        if not isinstance(s, int):
            raise TypeError("expected an int, got %r" % (type(s),))

        if not 0 <= s <= 65535:
            raise OverflowError("Not in range, %d" % s)

        self.write(self._packers["H"].pack(s))

    def read_short(self):
        """
        Reads a 2 byte integer from the stream.
        """
        return self._packers["h"].unpack(self.read(2))[0]

    def write_short(self, s):
        """
        Writes a 2 byte integer to the stream.

        @param s: 2 byte integer
        @type s: C{int}
        @raise TypeError: Unexpected type for int C{s}.
        @raise OverflowError: Not in range.
        """
        if not isinstance(s, int):
            raise TypeError("expected an int, got %r" % (type(s),))

        if not -32768 <= s <= 32767:
            raise OverflowError("Not in range, %d" % s)

        self.write(self._packers["h"].pack(s))

    def read_ulong(self):
        """
        Reads a 4 byte unsigned integer from the stream.
        """
        return self._packers["L"].unpack(self.read(4))[0]

    def write_ulong(self, l):
        """
        Writes a 4 byte unsigned integer to the stream.

        @param l: 4 byte unsigned integer
        @type l: C{int}
        @raise TypeError: Unexpected type for int C{l}.
        @raise OverflowError: Not in range.
        """
        if not isinstance(l, int):
            raise TypeError("expected an int, got %r" % (type(l),))

        if not 0 <= l <= 4294967295:
            raise OverflowError("Not in range, %d" % l)

        self.write(self._packers["L"].pack(l))

    def read_long(self):
        """
        Reads a 4 byte integer from the stream.
        """
        return self._packers["l"].unpack(self.read(4))[0]

    def write_long(self, l):
        """
        Writes a 4 byte integer to the stream.

        @param l: 4 byte integer
        @type l: C{int}
        @raise TypeError: Unexpected type for int C{l}.
        @raise OverflowError: Not in range.
        """
        if not isinstance(l, int):
            raise TypeError("expected an int, got %r" % (type(l),))

        if not -2147483648 <= l <= 2147483647:
            raise OverflowError("Not in range, %d" % l)

        self.write(self._packers["l"].pack(l))

    def read_24bit_uint(self):
        """
        Reads a 24 bit unsigned integer from the stream.

        @since: 0.4
        """
        if self._is_little_endian():
            order = (0, 8, 16)
        else:
            order = (16, 8, 0)

        n = 0
        for x in order:
            n += (self.read_uchar() << x)

        return n

    def write_24bit_uint(self, n):
        """
        Writes a 24 bit unsigned integer to the stream.

        @since: 0.4
        @param n: 24 bit unsigned integer
        @type n: C{int}
        @raise TypeError: Unexpected type for int C{n}.
        @raise OverflowError: Not in range.
        """
        if not isinstance(n, int):
            raise TypeError("expected an int, got %r" % (type(n),))

        if not 0 <= n <= 0xffffff:
            raise OverflowError("n is out of range")

        if self._is_little_endian():
            order = (0, 8, 16)
        else:
            order = (16, 8, 0)

        for x in order:
            self.write_uchar((n >> x) & 0xff)

    def read_24bit_int(self):
        """
        Reads a 24 bit integer from the stream.

        @since: 0.4
        """
        n = self.read_24bit_uint()

        if n & 0x800000 != 0:
            # the int is signed
            n -= 0x1000000

        return n

    def write_24bit_int(self, n):
        """
        Writes a 24 bit integer to the stream.

        @since: 0.4
        @param n: 24 bit integer
        @type n: C{int}
        @raise TypeError: Unexpected type for int C{n}.
        @raise OverflowError: Not in range.
        """
        if not isinstance(n, int):
            raise TypeError("expected an int, got %r" % (type(n),))

        if not -8388608 <= n <= 8388607:
            raise OverflowError("n is out of range")

        if n < 0:
            n += 0x1000000

        if self._is_little_endian():
            order = (0, 8, 16)
        else:
            order = (16, 8, 0)

        for x in order:
            self.write_uchar((n >> x) & 0xff)

    def read_double(self):
        """
        Reads an 8 byte float from the stream.
        """
        return self._packers["d"].unpack(self.read(8))[0]

    def write_double(self, d):
        """
        Writes an 8 byte float to the stream.

        @param d: 8 byte float
        @type d: C{float}
        @raise TypeError: Unexpected type for float C{d}.
        """
        if not isinstance(d, float):
            raise TypeError("expected a float, got %r" % (type(d),))

        self.write(self._packers["d"].pack(d))

    def read_float(self):
        """
        Reads a 4 byte float from the stream.
        """
        return self._packers["f"].unpack(self.read(4))[0]

    def write_float(self, f):
        """
        Writes a 4 byte float to the stream.

        @param f: 4 byte float
        @type f: C{float}
        @raise TypeError: Unexpected type for float C{f}.
        """
        if not isinstance(f, float):
            raise TypeError("expected a float, got %r" % (type(f),))

        self.write(self._packers["f"].pack(f))

    def read_utf8_string(self, length):
        """
        Reads a UTF-8 string from the stream.

        @rtype: Unicode string
        """
        return self.read(length).decode("utf-8")

    def write_utf8_string(self, u):
        """
        Writes a string to the stream.  If it is a Unicode object,
        it will be encoded in UTF-8; if it is a byte string, it will
        be written out as-is.

        @param u: string
        @raise TypeError: Unexpected type for C{u}
        """
        if isinstance(u, str):
            u = u.encode("utf-8")
        elif isinstance(u, bytearray):
            u = bytes(u)
        if not isinstance(u, bytes):
            raise TypeError("Expected a string, got %r" % (u,))
        self.write(u)



# Copyright (c) The PyAMF Project.
# See LICENSE.txt for details.

"""
Tests miniamf.util.imports

@since: 0.3.1
"""

import os.path
import sys
import unittest

from miniamf.util import imports


class InstalledTestCase(unittest.TestCase):
    """
    Tests to ensure that L{imports.finder} is installed in L{sys.meta_path}
    """

    def test_installed(self):
        f = imports.finder

        self.assertTrue(f in sys.meta_path)
        self.assertIdentical(sys.meta_path[0], f)


class ImportsTestCase(unittest.TestCase):
    def setUp(self):
        self.finder = imports.finder

        self._state = self.finder.__getstate__()

        path = os.path.join(os.path.dirname(__file__), 'imports')
        sys.path.insert(0, path)

    def tearDown(self):
        self.finder.__setstate__(self._state)

        del sys.path[0]
        self._clearModules('spam')

    def _clearModules(self, *args):
        for mod in args:
            for k, v in sys.modules.copy().items():
                if k.startswith(mod) or k == 'tests.%s' % (mod,):
                    del sys.modules[k]


class WhenImportedTestCase(ImportsTestCase):
    """
    Tests for L{imports.when_imported}
    """

    def setUp(self):
        ImportsTestCase.setUp(self)

        self.executed = False

    def _hook(self, module):
        self.executed = True

    def _check_module(self, mod):
        name = mod.__name__

        self.assertTrue(name in sys.modules)
        self.assertIdentical(sys.modules[name], mod)

    def test_import(self):
        imports.when_imported('spam', self._hook)

        self.assertFalse(self.executed)

        import spam

        self._check_module(spam)
        self.assertTrue(self.executed)

    def test_already_imported(self):
        import spam

        self.assertFalse(self.executed)

        imports.when_imported('spam', self._hook)

        self._check_module(spam)
        self.assertTrue(self.executed)

    def test_failed_hook(self):
        def h(mod):
            raise RuntimeError

        imports.when_imported('spam', h)

        try:
            import spam  # noqa
        except RuntimeError as e:
            pass
        except Exception as e:
            self.fail('expected RuntimeError, not %r' % (e,))
        else:
            self.fail('expected exception')

        self.assertFalse('spam' in self.finder.loaded_modules)



# Copyright (c) The PyAMF Project.
# See LICENSE.txt for details.


cdef class cBufferedByteStream:
    """
    The c version of BufferedByteStream.

    :todo: Need to spend some time marrying these two up.
    """
    cdef char endian
    cdef char *buffer
    cdef int closed
    cdef Py_ssize_t pos
    cdef Py_ssize_t size # total size of the alloc'd buffer
    cdef Py_ssize_t length
    cdef Py_ssize_t min_buf_size

    cdef inline int _init_buffer(self) except -1
    cpdef Py_ssize_t tell(self) except -1
    cdef int _actually_increase_buffer(self, Py_ssize_t size) except -1
    cdef inline int _increase_buffer(self, Py_ssize_t size) except -1
    cdef int write(self, char *buf, Py_ssize_t size) except -1
    cdef inline bint has_available(self, Py_ssize_t size) except -1
    cdef int read(self, char **buf, Py_ssize_t size) except -1
    cpdef bint at_eof(self) except -1
    cpdef Py_ssize_t remaining(self) except -1
    cpdef int seek(self, Py_ssize_t pos, int mode=*) except -1
    cpdef object getvalue(self)
    cdef Py_ssize_t peek(self, char **buf, Py_ssize_t size) except -1
    cpdef int truncate(self, Py_ssize_t size=?) except -1
    cpdef int consume(self) except -1
    cdef int unpack_int(self, unsigned int num_bytes, void *ret) except -1
    cdef int pack_int(self, int num_bytes, long x) except -1
    cdef int pack_uint(self, int num_bytes, unsigned long x) except -1
    cpdef unsigned char read_uchar(self) except? 0
    cpdef char read_char(self) except? 0
    cpdef unsigned short read_ushort(self) except? 0
    cpdef short read_short(self) except? 0
    cpdef unsigned long read_24bit_uint(self) except? 0
    cpdef long read_24bit_int(self) except? 0
    cpdef unsigned long read_ulong(self) except? 0
    cpdef long read_long(self) except? 0
    cpdef int write_uchar(self, unsigned char ret) except -1
    cpdef int write_char(self, char ret) except -1
    cpdef int write_ushort(self, unsigned short ret) except -1
    cpdef int write_short(self, short ret) except -1
    cpdef int write_ulong(self, unsigned long ret) except -1
    cpdef int write_long(self, long ret) except -1
    cpdef int write_24bit_uint(self, unsigned long ret) except -1
    cpdef int write_24bit_int(self, long ret) except -1
    cpdef object read_utf8_string(self, Py_ssize_t)
    cpdef int write_utf8_string(self, object obj) except -1
    cdef int read_double(self, double *obj) except -1
    cpdef int write_double(self, double val) except -1
    cdef int read_float(self, float *x) except -1
    cpdef int write_float(self, float c) except -1
    cpdef int append(self, object obj) except -1


cdef class BufferedByteStream(cBufferedByteStream):
    """
    A file like object that mimics StringIO with some extra helper methods.

    Provides the ability to read arbitrary data types from the underlying
    stream.
    """



# Copyright (c) The PyAMF Project.
# See LICENSE.txt for details.

"""
Adapter for the stdlib C{sets} module.

@since: 0.4
"""

import miniamf


def to_sorted_tuple(obj, encoder):
    return tuple(sorted(obj))


miniamf.add_type(frozenset, to_sorted_tuple)
miniamf.add_type(set, to_sorted_tuple)



# -*- encoding: utf-8 -*-
# Copyright (c) The PyAMF Project.
# See LICENSE.txt for details.

"""
General tests.

@since: 0.1.0
"""

from types import ModuleType
import unittest

import miniamf
from miniamf.tests.util import ClassCacheClearingTestCase, replace_dict, Spam


class ASObjectTestCase(unittest.TestCase):
    """
    I exercise all functionality relating to the L{ASObject<miniamf.ASObject>}
    class.
    """

    def test_init(self):
        bag = miniamf.ASObject(spam='eggs', baz='spam')

        self.assertEqual(bag, dict(spam='eggs', baz='spam'))
        self.assertEqual(bag.spam, 'eggs')
        self.assertEqual(bag.baz, 'spam')

    def test_eq(self):
        bag = miniamf.ASObject()

        self.assertEqual(bag, {})
        self.assertNotEqual(bag, {'spam': 'eggs'})

        bag2 = miniamf.ASObject()

        self.assertEqual(bag2, {})
        self.assertEqual(bag, bag2)
        self.assertNotEqual(bag, None)

    def test_setitem(self):
        bag = miniamf.ASObject()

        self.assertEqual(bag, {})

        bag['spam'] = 'eggs'

        self.assertEqual(bag.spam, 'eggs')

    def test_delitem(self):
        bag = miniamf.ASObject({'spam': 'eggs'})

        self.assertEqual(bag.spam, 'eggs')
        del bag['spam']

        self.assertRaises(AttributeError, lambda: bag.spam)

    def test_getitem(self):
        bag = miniamf.ASObject({'spam': 'eggs'})

        self.assertEqual(bag['spam'], 'eggs')

    def test_iter(self):
        bag = miniamf.ASObject({'spam': 'eggs'})

        x = []

        for k, v in bag.items():
            x.append((k, v))

        self.assertEqual(x, [('spam', 'eggs')])

    def test_hash(self):
        bag = miniamf.ASObject({'spam': 'eggs'})

        self.assertNotEqual(None, hash(bag))


class HelperTestCase(unittest.TestCase):
    """
    Tests all helper functions in C{miniamf.__init__}
    """

    def setUp(self):
        self.default_encoding = miniamf.DEFAULT_ENCODING

    def tearDown(self):
        miniamf.DEFAULT_ENCODING = self.default_encoding

    def test_get_decoder(self):
        self.assertRaises(ValueError, miniamf.get_decoder, 'spam')

        decoder = miniamf.get_decoder(miniamf.AMF0, stream=b'123', strict=True)
        self.assertEqual(decoder.stream.getvalue(), b'123')
        self.assertTrue(decoder.strict)

        decoder = miniamf.get_decoder(miniamf.AMF3, stream=b'456', strict=True)
        self.assertEqual(decoder.stream.getvalue(), b'456')
        self.assertTrue(decoder.strict)

    def test_get_encoder(self):
        miniamf.get_encoder(miniamf.AMF0)
        miniamf.get_encoder(miniamf.AMF3)
        self.assertRaises(ValueError, miniamf.get_encoder, b'spam')

        encoder = miniamf.get_encoder(miniamf.AMF0, stream=b'spam')
        self.assertEqual(encoder.stream.getvalue(), b'spam')
        self.assertFalse(encoder.strict)

        encoder = miniamf.get_encoder(miniamf.AMF3, stream=b'eggs')
        self.assertFalse(encoder.strict)

        encoder = miniamf.get_encoder(miniamf.AMF0, strict=True)
        self.assertTrue(encoder.strict)

        encoder = miniamf.get_encoder(miniamf.AMF3, strict=True)
        self.assertTrue(encoder.strict)

    def test_encode(self):
        self.assertEqual(
            miniamf.encode(u'connect', 1.0).getvalue(),
            b'\x06\x0fconnect\x05?\xf0\x00\x00\x00\x00\x00\x00'
        )

    def test_decode(self):
        self.assertEqual(
            list(miniamf.decode(
                b'\x06\x0fconnect\x05?\xf0\x00\x00\x00\x00\x00\x00')),
            [u'connect', 1.0]
        )

    def test_default_encoding(self):
        miniamf.DEFAULT_ENCODING = miniamf.AMF3

        x = miniamf.encode('foo').getvalue()

        self.assertEqual(x, b'\x06\x07foo')

        miniamf.DEFAULT_ENCODING = miniamf.AMF0

        x = miniamf.encode('foo').getvalue()

        self.assertEqual(x, b'\x02\x00\x03foo')


class UnregisterClassTestCase(ClassCacheClearingTestCase):
    def test_klass(self):
        alias = miniamf.register_class(Spam, 'spam.eggs')

        miniamf.unregister_class(Spam)
        self.assertTrue('spam.eggs' not in miniamf.CLASS_CACHE)
        self.assertTrue(Spam not in miniamf.CLASS_CACHE)
        self.assertTrue(alias not in miniamf.CLASS_CACHE)

    def test_alias(self):
        alias = miniamf.register_class(Spam, 'spam.eggs')

        miniamf.unregister_class('spam.eggs')
        self.assertTrue('spam.eggs' not in miniamf.CLASS_CACHE)
        self.assertTrue(alias not in miniamf.CLASS_CACHE)


class ClassLoaderTestCase(ClassCacheClearingTestCase):
    def test_register(self):
        self.assertTrue(chr not in miniamf.CLASS_LOADERS)
        miniamf.register_class_loader(chr)
        self.assertTrue(chr in miniamf.CLASS_LOADERS)

    def test_bad_register(self):
        self.assertRaises(TypeError, miniamf.register_class_loader, 1)
        miniamf.register_class_loader(ord)

    def test_unregister(self):
        self.assertTrue(chr not in miniamf.CLASS_LOADERS)
        miniamf.register_class_loader(chr)
        self.assertTrue(chr in miniamf.CLASS_LOADERS)

        miniamf.unregister_class_loader(chr)
        self.assertTrue(chr not in miniamf.CLASS_LOADERS)

        self.assertRaises(LookupError, miniamf.unregister_class_loader, chr)

    def test_load_class(self):
        def class_loader(x):
            self.assertEqual(x, 'spam.eggs')

            return Spam

        miniamf.register_class_loader(class_loader)

        self.assertTrue('spam.eggs' not in miniamf.CLASS_CACHE)
        miniamf.load_class('spam.eggs')
        self.assertTrue('spam.eggs' in miniamf.CLASS_CACHE)

    def test_load_unknown_class(self):
        def class_loader(x):
            return None

        miniamf.register_class_loader(class_loader)

        with self.assertRaises(miniamf.UnknownClassAlias):
            miniamf.load_class('spam.eggs')

    def test_load_class_by_alias(self):
        def class_loader(x):
            self.assertEqual(x, 'spam.eggs')
            return miniamf.ClassAlias(Spam, 'spam.eggs')

        miniamf.register_class_loader(class_loader)

        self.assertTrue('spam.eggs' not in miniamf.CLASS_CACHE)
        miniamf.load_class('spam.eggs')
        self.assertTrue('spam.eggs' in miniamf.CLASS_CACHE)

    def test_load_class_bad_return(self):
        def class_loader(x):
            return 'xyz'

        miniamf.register_class_loader(class_loader)

        self.assertRaises(TypeError, miniamf.load_class, 'spam.eggs')

    def test_load_class_by_module(self):
        miniamf.load_class('unittest.TestCase')

    def test_load_class_by_module_bad(self):
        with self.assertRaises(miniamf.UnknownClassAlias):
            miniamf.load_class('unittest.TestCase.')


class TypeMapTestCase(unittest.TestCase):
    def setUp(self):
        self.tm = miniamf.TYPE_MAP.copy()

        self.addCleanup(replace_dict, self.tm, miniamf.TYPE_MAP)

    def test_add_invalid(self):
        mod = ModuleType('spam')
        self.assertRaises(TypeError, miniamf.add_type, mod)
        self.assertRaises(TypeError, miniamf.add_type, {})
        self.assertRaises(TypeError, miniamf.add_type, 'spam')
        self.assertRaises(TypeError, miniamf.add_type, u'eggs')
        self.assertRaises(TypeError, miniamf.add_type, 1)
        self.assertRaises(TypeError, miniamf.add_type, 234234)
        self.assertRaises(TypeError, miniamf.add_type, 34.23)
        self.assertRaises(TypeError, miniamf.add_type, None)
        self.assertRaises(TypeError, miniamf.add_type, object())

        class A:
            pass

        self.assertRaises(TypeError, miniamf.add_type, A())

    def test_add_same(self):
        miniamf.add_type(chr)
        self.assertRaises(KeyError, miniamf.add_type, chr)

    def test_add_class(self):
        class A:
            pass

        class B(object):
            pass

        miniamf.add_type(A)
        self.assertTrue(A in miniamf.TYPE_MAP)

        miniamf.add_type(B)
        self.assertTrue(B in miniamf.TYPE_MAP)

    def test_add_callable(self):
        td = miniamf.add_type(ord)

        self.assertTrue(ord in miniamf.TYPE_MAP)
        self.assertTrue(td in miniamf.TYPE_MAP.values())

    def test_add_multiple(self):
        td = miniamf.add_type((chr,))

        class A(object):
            pass

        class B(object):
            pass

        class C(object):
            pass

        td = miniamf.add_type([A, B, C])
        self.assertEqual(td, miniamf.get_type([A, B, C]))

    def test_get_type(self):
        self.assertRaises(KeyError, miniamf.get_type, chr)
        td = miniamf.add_type((chr,))
        self.assertRaises(KeyError, miniamf.get_type, chr)

        td2 = miniamf.get_type((chr, ))
        self.assertEqual(td, td2)

        td2 = miniamf.get_type([chr, ])
        self.assertEqual(td, td2)

    def test_remove(self):
        self.assertRaises(KeyError, miniamf.remove_type, chr)
        td = miniamf.add_type((chr,))

        self.assertRaises(KeyError, miniamf.remove_type, chr)
        td2 = miniamf.remove_type((chr,))

        self.assertEqual(td, td2)


class ErrorClassMapTestCase(unittest.TestCase):
    """
    I test all functionality related to manipulating L{miniamf.ERROR_CLASS_MAP}
    """

    def setUp(self):
        self.map_copy = miniamf.ERROR_CLASS_MAP.copy()
        self.addCleanup(replace_dict, self.map_copy, miniamf.ERROR_CLASS_MAP)

    def test_add(self):
        class A:
            pass

        class B(Exception):
            pass

        self.assertRaises(TypeError, miniamf.add_error_class, None, 'a')

        # class A does not sub-class Exception
        self.assertRaises(TypeError, miniamf.add_error_class, A, 'a')

        miniamf.add_error_class(B, 'b')
        self.assertEqual(miniamf.ERROR_CLASS_MAP['b'], B)

        miniamf.add_error_class(B, 'a')
        self.assertEqual(miniamf.ERROR_CLASS_MAP['a'], B)

        class C(Exception):
            pass

        self.assertRaises(ValueError, miniamf.add_error_class, C, 'b')

    def test_remove(self):
        class B(Exception):
            pass

        miniamf.ERROR_CLASS_MAP['abc'] = B

        self.assertRaises(TypeError, miniamf.remove_error_class, None)

        miniamf.remove_error_class('abc')
        self.assertFalse('abc' in miniamf.ERROR_CLASS_MAP)
        self.assertRaises(KeyError, miniamf.ERROR_CLASS_MAP.__getitem__, 'abc')

        miniamf.ERROR_CLASS_MAP['abc'] = B

        miniamf.remove_error_class(B)

        self.assertRaises(KeyError, miniamf.ERROR_CLASS_MAP.__getitem__, 'abc')
        self.assertRaises(ValueError, miniamf.remove_error_class, B)
        self.assertRaises(ValueError, miniamf.remove_error_class, 'abc')


class DummyAlias(miniamf.ClassAlias):
    pass


class RegisterAliasTypeTestCase(unittest.TestCase):
    def setUp(self):
        self.old_aliases = miniamf.ALIAS_TYPES.copy()
        self.addCleanup(replace_dict, self.old_aliases, miniamf.ALIAS_TYPES)

    def test_bad_klass(self):
        self.assertRaises(TypeError, miniamf.register_alias_type, 1)

    def test_subclass(self):
        self.assertFalse(issubclass(self.__class__, miniamf.ClassAlias))
        with self.assertRaises(ValueError):
            miniamf.register_alias_type(self.__class__)

    def test_no_args(self):
        self.assertTrue(issubclass(DummyAlias, miniamf.ClassAlias))
        self.assertRaises(ValueError, miniamf.register_alias_type, DummyAlias)

    def test_type_args(self):
        self.assertTrue(issubclass(DummyAlias, miniamf.ClassAlias))
        self.assertRaises(TypeError,
                          miniamf.register_alias_type, DummyAlias, 1)

    def test_single(self):
        class A(object):
            pass

        miniamf.register_alias_type(DummyAlias, A)

        self.assertTrue(DummyAlias in miniamf.ALIAS_TYPES)
        self.assertEqual(miniamf.ALIAS_TYPES[DummyAlias], (A,))

    def test_multiple(self):
        class A(object):
            pass

        class B(object):
            pass

        with self.assertRaises(TypeError):
            miniamf.register_alias_type(DummyAlias, A, 'hello')

        miniamf.register_alias_type(DummyAlias, A, B)
        self.assertTrue(DummyAlias in miniamf.ALIAS_TYPES)
        self.assertEqual(miniamf.ALIAS_TYPES[DummyAlias], (A, B))

    def test_duplicate(self):
        class A(object):
            pass

        miniamf.register_alias_type(DummyAlias, A)

        with self.assertRaises(RuntimeError):
            miniamf.register_alias_type(DummyAlias, A)

    def test_unregister(self):
        """
        Tests for L{miniamf.unregister_alias_type}
        """
        class A(object):
            pass

        self.assertFalse(DummyAlias in miniamf.ALIAS_TYPES)
        self.assertEqual(miniamf.unregister_alias_type(A), None)

        miniamf.register_alias_type(DummyAlias, A)

        self.assertTrue(DummyAlias in miniamf.ALIAS_TYPES)
        self.assertEqual(miniamf.unregister_alias_type(DummyAlias), (A,))


class TypedObjectTestCase(unittest.TestCase):
    def test_externalised(self):
        o = miniamf.TypedObject(None)

        self.assertRaises(miniamf.DecodeError, o.__readamf__, None)
        self.assertRaises(miniamf.EncodeError, o.__writeamf__, None)

    def test_alias(self):
        class Foo:
            pass

        alias = miniamf.TypedObjectClassAlias(Foo, 'bar')

        self.assertEqual(alias.klass, miniamf.TypedObject)
        self.assertNotEqual(alias.klass, Foo)


class PackageTestCase(ClassCacheClearingTestCase):
    """
    Tests for L{miniamf.register_package}
    """

    class NewType(object):
        pass

    class ClassicType:
        pass

    def setUp(self):
        ClassCacheClearingTestCase.setUp(self)

        self.module = ModuleType("foo")

        self.module.Classic = self.ClassicType
        self.module.New = self.NewType
        self.module.b = b'binary'
        self.module.i = 12323
        self.module.f = 345.234
        self.module.u = u"Unicöde"
        self.module.l = ["list", "of", "junk"]
        self.module.d = {"foo": "bar", "baz": "gak"}
        self.module.obj = object()
        self.module.mod = self.module
        self.module.lam = lambda _: None

        self.NewType.__module__ = "foo"
        self.ClassicType.__module__ = "foo"

        self.spam_module = Spam.__module__
        Spam.__module__ = "foo"

        self.names = (self.module.__name__,)

    def tearDown(self):
        ClassCacheClearingTestCase.tearDown(self)

        Spam.__module__ = self.spam_module

        self.module.__name__ = self.names

    def check_module(self, r, base_package):
        self.assertEqual(len(r), 2)

        for c in [self.NewType, self.ClassicType]:
            alias = r[c]

            self.assertTrue(isinstance(alias, miniamf.ClassAlias))
            self.assertEqual(alias.klass, c)
            self.assertEqual(alias.alias, base_package + c.__name__)

    def test_module(self):
        r = miniamf.register_package(self.module, 'com.example')
        self.check_module(r, 'com.example.')

    def test_all(self):
        self.module.Spam = Spam

        self.module.__all__ = ['Classic', 'New']

        r = miniamf.register_package(self.module, 'com.example')
        self.check_module(r, 'com.example.')

    def test_ignore(self):
        self.module.Spam = Spam

        r = miniamf.register_package(self.module, 'com.example',
                                     ignore=['Spam'])
        self.check_module(r, 'com.example.')

    def test_separator(self):
        r = miniamf.register_package(self.module, 'com.example', separator='/')

        self.ClassicType.__module__ = 'com.example'
        self.NewType.__module__ = 'com.example'
        self.check_module(r, 'com.example/')

    def test_name(self):
        self.module.__name__ = 'spam.eggs'
        self.ClassicType.__module__ = 'spam.eggs'
        self.NewType.__module__ = 'spam.eggs'

        r = miniamf.register_package(self.module)
        self.check_module(r, 'spam.eggs.')

    def test_dict(self):
        """
        @see: #585
        """
        d = dict()
        d['Spam'] = Spam

        r = miniamf.register_package(d, 'com.example', strict=False)

        self.assertEqual(len(r), 1)

        alias = r[Spam]

        self.assertTrue(isinstance(alias, miniamf.ClassAlias))
        self.assertEqual(alias.klass, Spam)
        self.assertEqual(alias.alias, 'com.example.Spam')

    def test_odd(self):
        self.assertRaises(TypeError, miniamf.register_package, object())
        self.assertRaises(TypeError, miniamf.register_package, 1)
        self.assertRaises(TypeError, miniamf.register_package, 1.2)
        self.assertRaises(TypeError, miniamf.register_package, 23897492834)
        self.assertRaises(TypeError, miniamf.register_package, [])
        self.assertRaises(TypeError, miniamf.register_package, b'')
        self.assertRaises(TypeError, miniamf.register_package, u'')

    def test_strict(self):
        self.module.Spam = Spam

        Spam.__module__ = self.spam_module

        r = miniamf.register_package(self.module, 'com.example', strict=True)
        self.check_module(r, 'com.example.')

    def test_not_strict(self):
        self.module.Spam = Spam

        Spam.__module__ = self.spam_module

        r = miniamf.register_package(self.module, 'com.example', strict=False)

        self.assertEqual(len(r), 3)

        for c in [self.NewType, self.ClassicType, Spam]:
            alias = r[c]

            self.assertTrue(isinstance(alias, miniamf.ClassAlias))
            self.assertEqual(alias.klass, c)
            self.assertEqual(alias.alias, 'com.example.' + c.__name__)

    def test_list(self):
        class Foo:
            pass

        class Bar:
            pass

        ret = miniamf.register_package([Foo, Bar], 'spam.eggs')

        self.assertEqual(len(ret), 2)

        for c in [Foo, Bar]:
            alias = ret[c]

            self.assertTrue(isinstance(alias, miniamf.ClassAlias))
            self.assertEqual(alias.klass, c)
            self.assertEqual(alias.alias, 'spam.eggs.' + c.__name__)


class UndefinedTestCase(unittest.TestCase):
    """
    Tests for L{miniamf.Undefined}
    """

    def test_none(self):
        """
        L{miniamf.Undefined} is not referentially identical to C{None}.
        """
        self.assertFalse(miniamf.Undefined is None)

    def test_non_zero(self):
        """
        Truth test for L{miniamf.Undefined} == C{False}.
        """
        self.assertFalse(miniamf.Undefined)


class TestAMF0Codecs(unittest.TestCase):
    """
    Tests for getting encoder/decoder for AMF0 with extension support.
    """

    def test_default_decoder(self):
        """
        If the extension is available, it must be returned by default.
        """
        try:
            from miniamf._accel import amf0
        except ImportError:
            from miniamf import amf0

        decoder = miniamf.get_decoder(miniamf.AMF0)

        self.assertIsInstance(decoder, amf0.Decoder)

    def test_ext_decoder(self):
        """
        With `use_ext=True` specified, the extension must be returned.
        """
        try:
            from miniamf._accel import amf0
        except ImportError:
            self.skipTest('amf0 extension not available')

        decoder = miniamf.get_decoder(miniamf.AMF0, use_ext=True)

        self.assertIsInstance(decoder, amf0.Decoder)

    def test_pure_decoder(self):
        """
        With `use_ext=False` specified, the extension must NOT be returned.
        """
        from miniamf import amf0

        decoder = miniamf.get_decoder(miniamf.AMF0, use_ext=False)

        self.assertIsInstance(decoder, amf0.Decoder)

    def test_default_encoder(self):
        """
        If the extension is available, it must be returned by default.
        """
        try:
            from miniamf._accel import amf0
        except ImportError:
            from miniamf import amf0

        encoder = miniamf.get_encoder(miniamf.AMF0)

        self.assertIsInstance(encoder, amf0.Encoder)

    def test_ext_encoder(self):
        """
        With `use_ext=True` specified, the extension must be returned.
        """
        try:
            from miniamf._accel import amf0
        except ImportError:
            self.skipTest('amf0 extension not available')

        encoder = miniamf.get_encoder(miniamf.AMF0, use_ext=True)

        self.assertIsInstance(encoder, amf0.Encoder)

    def test_pure_encoder(self):
        """
        With `use_ext=False` specified, the extension must NOT be returned.
        """
        from miniamf import amf0

        encoder = miniamf.get_encoder(miniamf.AMF0, use_ext=False)

        self.assertIsInstance(encoder, amf0.Encoder)


class TestAMF3Codecs(unittest.TestCase):
    """
    Tests for getting encoder/decoder for amf3 with extension support.
    """

    def test_default_decoder(self):
        """
        If the extension is available, it must be returned by default.
        """
        try:
            from miniamf._accel import amf3
        except ImportError:
            from miniamf import amf3

        decoder = miniamf.get_decoder(miniamf.AMF3)

        self.assertIsInstance(decoder, amf3.Decoder)

    def test_ext_decoder(self):
        """
        With `use_ext=True` specified, the extension must be returned.
        """
        try:
            from miniamf._accel import amf3
        except ImportError:
            self.skipTest('amf3 extension not available')

        decoder = miniamf.get_decoder(miniamf.AMF3, use_ext=True)

        self.assertIsInstance(decoder, amf3.Decoder)

    def test_pure_decoder(self):
        """
        With `use_ext=False` specified, the extension must NOT be returned.
        """
        from miniamf import amf3

        decoder = miniamf.get_decoder(miniamf.AMF3, use_ext=False)

        self.assertIsInstance(decoder, amf3.Decoder)

    def test_default_encoder(self):
        """
        If the extension is available, it must be returned by default.
        """
        try:
            from miniamf._accel import amf3
        except ImportError:
            from miniamf import amf3

        encoder = miniamf.get_encoder(miniamf.AMF3)

        self.assertIsInstance(encoder, amf3.Encoder)

    def test_ext_encoder(self):
        """
        With `use_ext=True` specified, the extension must be returned.
        """
        try:
            from miniamf._accel import amf3
        except ImportError:
            self.skipTest('amf3 extension not available')

        encoder = miniamf.get_encoder(miniamf.AMF3, use_ext=True)

        self.assertIsInstance(encoder, amf3.Encoder)

    def test_pure_encoder(self):
        """
        With `use_ext=False` specified, the extension must NOT be returned.
        """
        from miniamf import amf3

        encoder = miniamf.get_encoder(miniamf.AMF3, use_ext=False)

        self.assertIsInstance(encoder, amf3.Encoder)



class User(object):
    def __init__(self, name, password):
        self.name = name
        self.password = password

class Permission(object):
    def __init__(self, type):
        self.type = type



# -*- coding: utf-8 -*-
#
# Copyright (c) The PyAMF Project.
# See LICENSE.txt for details.

"""
Tests for AMF0 Implementation.

@since: 0.1.0
"""

import datetime
import math
import unittest

import miniamf
from miniamf import amf0, util, xml
from miniamf.tests.util import EncoderMixIn, DecoderMixIn, ClassCacheClearingTestCase, Spam, ClassicSpam


class TypesTestCase(unittest.TestCase):
    """
    Tests the type mappings.
    """

    def test_types(self):
        self.assertEqual(amf0.TYPE_NUMBER, b'\x00')
        self.assertEqual(amf0.TYPE_BOOL, b'\x01')
        self.assertEqual(amf0.TYPE_STRING, b'\x02')
        self.assertEqual(amf0.TYPE_OBJECT, b'\x03')
        self.assertEqual(amf0.TYPE_MOVIECLIP, b'\x04')
        self.assertEqual(amf0.TYPE_NULL, b'\x05')
        self.assertEqual(amf0.TYPE_UNDEFINED, b'\x06')
        self.assertEqual(amf0.TYPE_REFERENCE, b'\x07')
        self.assertEqual(amf0.TYPE_MIXEDARRAY, b'\x08')
        self.assertEqual(amf0.TYPE_OBJECTTERM, b'\x09')
        self.assertEqual(amf0.TYPE_ARRAY, b'\x0a')
        self.assertEqual(amf0.TYPE_DATE, b'\x0b')
        self.assertEqual(amf0.TYPE_LONGSTRING, b'\x0c')
        self.assertEqual(amf0.TYPE_UNSUPPORTED, b'\x0d')
        self.assertEqual(amf0.TYPE_RECORDSET, b'\x0e')
        self.assertEqual(amf0.TYPE_XML, b'\x0f')
        self.assertEqual(amf0.TYPE_TYPEDOBJECT, b'\x10')
        self.assertEqual(amf0.TYPE_AMF3, b'\x11')


class EncoderTestCase(ClassCacheClearingTestCase, EncoderMixIn):
    """
    Tests the output from the AMF0 L{Encoder<miniamf.amf0.Encoder>} class.
    """

    amf_type = miniamf.AMF0

    def setUp(self):
        ClassCacheClearingTestCase.setUp(self)
        EncoderMixIn.setUp(self)

    def test_number(self):
        self.assertEncoded(0, b'\x00\x00\x00\x00\x00\x00\x00\x00\x00')
        self.assertEncoded(0.2, b'\x00\x3f\xc9\x99\x99\x99\x99\x99\x9a')
        self.assertEncoded(1, b'\x00\x3f\xf0\x00\x00\x00\x00\x00\x00')
        self.assertEncoded(42, b'\x00\x40\x45\x00\x00\x00\x00\x00\x00')
        self.assertEncoded(-123, b'\x00\xc0\x5e\xc0\x00\x00\x00\x00\x00')
        self.assertEncoded(1.23456789, b'\x00\x3f\xf3\xc0\xca\x42\x83\xde\x1b')

    def test_boolean(self):
        self.assertEncoded(True, b'\x01\x01')
        self.assertEncoded(False, b'\x01\x00')

    def test_string(self):
        self.assertEncoded('', b'\x02\x00\x00')
        self.assertEncoded('hello', b'\x02\x00\x05hello')
        # Unicode taken from http://www.columbia.edu/kermit/utf8.html
        self.assertEncoded(
            u'ᚠᛇᚻ', b'\x02\x00\t\xe1\x9a\xa0\xe1\x9b\x87\xe1\x9a\xbb'
        )

    def test_null(self):
        self.assertEncoded(None, b'\x05')

    def test_undefined(self):
        self.assertEncoded(miniamf.Undefined, b'\x06')

    def test_list(self):
        self.assertEncoded([], b'\x0a\x00\x00\x00\x00')
        self.assertEncoded(
            [1, 2, 3],
            b'\x0a\x00\x00\x00\x03\x00\x3f\xf0\x00\x00\x00\x00\x00\x00\x00\x40'
            b'\x00\x00\x00\x00\x00\x00\x00\x00\x40\x08\x00\x00\x00\x00\x00\x00'
        )
        self.assertEncoded(
            (1, 2, 3),
            b'\x0a\x00\x00\x00\x03\x00\x3f\xf0\x00\x00\x00\x00\x00\x00\x00\x40'
            b'\x00\x00\x00\x00\x00\x00\x00\x00\x40\x08\x00\x00\x00\x00\x00\x00'
        )

    def test_list_references(self):
        x = []

        self.assertEqual(self.encode(x, x), b'\n\x00\x00\x00\x00\x07\x00\x00')

    def test_longstring(self):
        s = b'a' * 65537

        self.assertEncoded(s, b'\x0c\x00\x01\x00\x01' + s)

    def test_dict(self):
        self.assertEncoded({'a': b'a'},
                           b'\x03\x00\x01a\x02\x00\x01a\x00\x00\t')

        self.assertEncoded({12: True, 42: "Testing"}, b'\x03', (
            b'\x00\x0242\x02\x00\x07Testing',
            b'\x00\x0212\x01\x01'
        ), b'\x00\x00\t')

    def test_mixed_array(self):
        d = miniamf.MixedArray(a=1, b=2, c=3)

        bytes = (b'\x08\x00\x00\x00\x00', (
            b'\x00\x01a\x00?\xf0\x00\x00\x00\x00\x00\x00',
            b'\x00\x01c\x00@\x08\x00\x00\x00\x00\x00\x00',
            b'\x00\x01b\x00@\x00\x00\x00\x00\x00\x00\x00'
        ), b'\x00\x00\t')

        self.assertEncoded(d, bytes)

        # test the reference
        self.assertEqual(self.encode(d), b'\x07\x00\x00')

    def test_date(self):
        self.assertEncoded(
            datetime.datetime(2005, 3, 18, 1, 58, 31),
            b'\x0bBp+6!\x15\x80\x00\x00\x00'
        )
        self.assertEncoded(
            datetime.date(2003, 12, 1),
            b'\x0bBo%\xe2\xb2\x80\x00\x00\x00\x00'
        )
        self.assertEncoded(
            datetime.datetime(2009, 3, 8, 23, 30, 47, 770122),
            b'\x0bBq\xfe\x86\xca5\xa1\xf4\x00\x00'
        )

        self.assertRaises(miniamf.EncodeError, self.encode,
                          datetime.time(22, 3))

    def test_xml(self):
        blob = b'<a><b>hello world</b></a>'

        self.assertEncoded(
            xml.fromstring(blob),
            b'\x0f\x00\x00\x00\x19' + blob
        )

    def test_xml_references(self):
        blob = b'<a><b>hello world</b></a>'
        x = xml.fromstring(blob)

        self.assertEncoded(
            [x, x],
            b'\n\x00\x00\x00\x02' + (b'\x0f\x00\x00\x00\x19' + blob) * 2
        )

    def test_object(self):
        self.assertEncoded(
            {'a': 'b'},
            b'\x03\x00\x01a\x02\x00\x01b\x00\x00\x09'
        )

    def test_force_amf3(self):
        alias = miniamf.register_class(Spam, 'spam.eggs')
        alias.amf3 = True

        x = Spam()
        x.x = 'y'

        self.assertEncoded(x, b'\x11\n\x0b\x13spam.eggs\x03x\x06\x03y\x01')

    def test_typed_object(self):
        miniamf.register_class(Spam, alias='org.miniamf.spam')

        x = Spam()
        x.baz = 'hello'

        self.assertEncoded(
            x,
            b'\x10\x00\x10org.miniamf.spam'
            b'\x00\x03baz\x02\x00\x05hello\x00\x00\t'
        )

    def test_complex_list(self):
        self.assertEncoded(
            [[1.0]],
            b'\x0a\x00\x00\x00\x01\x0a\x00\x00\x00\x01\x00\x3f\xf0\x00\x00\x00'
            b'\x00\x00\x00'
        )

        self.assertEncoded(
            [['test', 'test', 'test', 'test']],
            b'\x0a\x00\x00\x00\x01\x0a\x00\x00\x00\x04' +
            (b'\x02\x00\x04test' * 4)
        )

        x = {'a': 'spam', 'b': 'eggs'}

        self.assertEncoded(
            [[x, x]],
            b'\n\x00\x00\x00\x01\n\x00\x00\x00\x02\x03\x00\x01a\x02\x00'
            b'\x04spam\x00\x01b\x02\x00\x04eggs\x00\x00\t\x07\x00\x02'
        )

    def test_amf3(self):
        self.encoder.use_amf3 = True

        o = Spam()

        self.assertEncoded(o, b'\x11\n\x0b\x01\x01')

    def test_anonymous(self):
        miniamf.register_class(Spam)

        x = Spam()
        x.spam = 'eggs'
        x.hello = 'world'

        self.assertEncoded(
            x,
            b'\x03',
            (b'\x00\x05hello\x02\x00\x05world',
             b'\x00\x04spam\x02\x00\x04eggs'),
            b'\x00\x00\t'
        )

    def test_dynamic(self):
        x = Spam()

        x.foo = 'bar'
        x.hello = 'world'

        alias = miniamf.register_class(Spam)

        alias.exclude_attrs = ['foo']

        alias.compile()

        self.assertTrue(alias.dynamic)

        self.assertEncoded(x, b'\x03\x00\x05hello\x02\x00\x05world\x00\x00\t')

    def test_dynamic_static(self):
        x = Spam()

        x.foo = 'bar'
        x.hello = 'world'

        alias = miniamf.register_class(Spam)

        alias.static_attrs = ['hello']
        alias.compile()

        self.assertTrue(alias.dynamic)

        self.assertEncoded(
            x,
            b'\x03',
            (b'\x00\x05hello\x02\x00\x05world', b'\x00\x03foo\x02\x00\x03bar'),
            b'\x00\x00\t'
        )

    def test_dynamic_registered(self):
        x = Spam()

        x.foo = 'bar'
        x.hello = 'world'

        alias = miniamf.register_class(Spam, 'x')

        alias.exclude_attrs = ['foo']

        alias.compile()

        self.assertTrue(alias.dynamic)

        self.assertEncoded(
            x,
            b'\x10\x00\x01x',
            b'\x00\x05hello\x02\x00\x05world',
            b'\x00\x00\t'
        )

    def test_custom_type(self):
        def write_as_list(list_interface_obj, encoder):
            list_interface_obj.ran = True
            self.assertEqual(id(encoder), id(self.encoder))

            return list(list_interface_obj)

        class ListWrapper(object):
            ran = False

            def __iter__(self):
                return iter([1, 2, 3])

        miniamf.add_type(ListWrapper, write_as_list)
        x = ListWrapper()

        self.encoder.writeElement(x)
        self.assertEqual(x.ran, True)

        self.assertEqual(
            self.buf.getvalue(),
            b'\n\x00\x00\x00\x03\x00?\xf0\x00\x00\x00\x00\x00\x00\x00@\x00\x00'
            b'\x00\x00\x00\x00\x00\x00@\x08\x00\x00\x00\x00\x00\x00'
        )

    def test_old_style_classes(self):
        class Person:
            pass

        miniamf.register_class(Person, 'spam.eggs.Person')

        u = Person()
        u.family_name = 'Doe'
        u.given_name = 'Jane'

        self.encoder.writeElement(u)

        self.assertEncoded(u, b'\x10\x00\x10spam.eggs.Person', (
            b'\x00\x0bfamily_name\x02\x00\x03Doe',
            b'\x00\ngiven_name\x02\x00\x04Jane'
        ), b'\x00\x00\t')

    def test_slots(self):
        class Person(object):
            __slots__ = ('family_name', 'given_name')

        u = Person()
        u.family_name = 'Doe'
        u.given_name = 'Jane'

        self.assertEncoded(u, b'\x03', (
            b'\x00\x0bfamily_name\x02\x00\x03Doe',
            b'\x00\ngiven_name\x02\x00\x04Jane'
        ), b'\x00\x00\t')

    def test_slots_registered(self):
        class Person(object):
            __slots__ = ('family_name', 'given_name')

        u = Person()
        u.family_name = 'Doe'
        u.given_name = 'Jane'

        miniamf.register_class(Person, 'spam.eggs.Person')

        self.assertEncoded(u, b'\x10\x00\x10spam.eggs.Person', (
            b'\x00\x0bfamily_name\x02\x00\x03Doe',
            b'\x00\ngiven_name\x02\x00\x04Jane'
        ), b'\x00\x00\t')

    def test_elementtree_tag(self):
        """
        Pretend to look like an ElementTree object to try to fool Mini-AMF into
        encoding an xml type.
        """
        class NotAnElement(object):
            def items(self):
                return []

            def __iter__(self):
                return iter([])

        foo = NotAnElement()
        foo.tag = 'foo'
        foo.text = 'bar'
        foo.tail = None

        self.assertEncoded(
            foo,
            b'\x03', (
                b'\x00\x04text\x02\x00\x03bar',
                b'\x00\x04tail\x05',
                b'\x00\x03tag\x02\x00\x03foo',
            ),
            b'\x00\x00\t'
        )

    def test_funcs(self):
        def x():
            pass

        for i in (chr, self.assertRaises, lambda x: x, miniamf):
            self.assertRaises(miniamf.EncodeError,
                              self.encoder.writeElement, i)

    def test_external_subclassed_list(self):
        class L(list):
            class __amf__:
                external = True

            def __readamf__(self, o):
                pass

            def __writeamf__(self, o):
                pass

        miniamf.register_class(L, 'a')

        a = L()

        a.append('foo')
        a.append('bar')

        self.assertEncoded(a, b'\x10\x00\x01a\x00\x00\t')

    def test_nonexternal_subclassed_list(self):
        class L(list):
            pass

        miniamf.register_class(L, 'a')

        a = L()

        a.append('foo')
        a.append('bar')

        self.assertEncoded(
            a, b'\n\x00\x00\x00\x02\x02\x00\x03foo\x02\x00\x03bar'
        )

    def test_amf3_xml(self):
        self.encoder.use_amf3 = True
        blob = '<root><sections><section /><section /></sections></root>'

        blob = xml.tostring(xml.fromstring(blob))

        bytes = self.encode(xml.fromstring(blob))

        buf = util.BufferedByteStream(bytes)

        self.assertEqual(buf.read_uchar(), 17)
        self.assertEqual(buf.read_uchar(), 11)
        self.assertEqual(buf.read_uchar() >> 1, buf.remaining())
        self.assertEqual(buf.read(), blob)

    def test_use_amf3(self):
        self.encoder.use_amf3 = True

        x = {'foo': 'bar', 'baz': 'gak'}

        self.assertEncoded(
            x,
            b'\x11\n\x0b',
            (b'\x01\x07baz\x06\x07gak', b'\x07foo\x06\x07bar\x01')
        )

    def test_static_attrs(self):
        class Foo(object):
            class __amf__:
                static = ('foo', 'bar')

        miniamf.register_class(Foo)

        x = Foo()
        x.foo = 'baz'
        x.bar = 'gak'

        self.assertEncoded(
            x,
            b'\x03',
            (b'\x00\x03bar\x02\x00\x03gak', b'\x00\x03foo\x02\x00\x03baz'),
            b'\x00\x00\t'
        )

    def test_class(self):
        class Classic:
            pass

        class New(object):
            pass

        with self.assertRaises(miniamf.EncodeError):
            self.encoder.writeElement(Classic)

        with self.assertRaises(miniamf.EncodeError):
            self.encoder.writeElement(New)

    def test_timezone(self):
        d = datetime.datetime(2009, 9, 24, 14, 23, 23)
        self.encoder.timezone_offset = datetime.timedelta(hours=-5)

        self.assertEncoded(d, b'\x0bBr>\xd8\x1f\xff\x80\x00\x00\x00')

    def test_generators(self):
        def foo():
            yield [1, 2, 3]
            yield b'\xff'
            yield miniamf.Undefined

        self.assertEncoded(
            foo(),
            b'\n\x00\x00\x00\x03\x00?\xf0\x00\x00\x00\x00\x00\x00\x00'
            b'@\x00\x00\x00\x00\x00\x00\x00\x00@\x08\x00\x00\x00\x00'
            b'\x00\x00\x02\x00\x01\xff\x06'
        )

    def test_iterate(self):
        self.assertRaises(StopIteration, lambda: next(self.encoder))

        self.encoder.send('')
        self.encoder.send('hello')
        self.encoder.send(u'ƒøø')

        self.assertEqual(next(self.encoder), b'\x02\x00\x00')
        self.assertEqual(next(self.encoder), b'\x02\x00\x05hello')
        self.assertEqual(
            next(self.encoder),
            b'\x02\x00\x06\xc6\x92\xc3\xb8\xc3\xb8'
        )

        self.assertRaises(StopIteration, lambda: next(self.encoder))

        self.assertIdentical(iter(self.encoder), self.encoder)
        self.assertEqual(
            self.buf.getvalue(),
            b'\x02\x00\x00\x02\x00\x05hello'
            b'\x02\x00\x06\xc6\x92\xc3\xb8\xc3\xb8'
        )

    def test_subclassed_tuple(self):
        """
        A subclassed tuple must encode an AMF list.

        @see: #830
        """
        class Foo(tuple):
            pass

        x = Foo([1, 2])

        self.encoder.send(x)

        self.assertEqual(
            next(self.encoder),
            b'\n\x00\x00\x00\x02\x00?\xf0\x00\x00\x00\x00\x00\x00\x00@\x00\x00'
            b'\x00\x00\x00\x00\x00'
        )


class DecoderTestCase(ClassCacheClearingTestCase, DecoderMixIn):
    """
    Tests the output from the AMF0 L{Decoder<miniamf.amf0.Decoder>} class.
    """

    amf_type = miniamf.AMF0

    def setUp(self):
        ClassCacheClearingTestCase.setUp(self)
        DecoderMixIn.setUp(self)

    def test_undefined(self):
        self.assertDecoded(miniamf.Undefined, b'\x06')

    def test_number(self):
        self.assertDecoded(0, b'\x00\x00\x00\x00\x00\x00\x00\x00\x00')
        self.assertDecoded(0.2, b'\x00\x3f\xc9\x99\x99\x99\x99\x99\x9a')
        self.assertDecoded(1, b'\x00\x3f\xf0\x00\x00\x00\x00\x00\x00')
        self.assertDecoded(42, b'\x00\x40\x45\x00\x00\x00\x00\x00\x00')
        self.assertDecoded(-123, b'\x00\xc0\x5e\xc0\x00\x00\x00\x00\x00')
        self.assertDecoded(1.23456789, b'\x00\x3f\xf3\xc0\xca\x42\x83\xde\x1b')

    def test_number_types(self):
        nr_types = [
            (b'\x00\x00\x00\x00\x00\x00\x00\x00\x00', int),
            (b'\x00\x3f\xc9\x99\x99\x99\x99\x99\x9a', float),
            (b'\x00\x3f\xf0\x00\x00\x00\x00\x00\x00', int),
            (b'\x00\x40\x45\x00\x00\x00\x00\x00\x00', int),
            (b'\x00\xc0\x5e\xc0\x00\x00\x00\x00\x00', int),
            (b'\x00\x3f\xf3\xc0\xca\x42\x83\xde\x1b', float),
            (b'\x00\xff\xf8\x00\x00\x00\x00\x00\x00', float),  # nan
            (b'\x00\xff\xf0\x00\x00\x00\x00\x00\x00', float),  # -inf
            (b'\x00\x7f\xf0\x00\x00\x00\x00\x00\x00', float),  # inf
        ]

        for t in nr_types:
            bytes, expected_type = t
            self.buf.truncate()
            self.buf.seek(0, 0)
            self.buf.append(bytes)
            self.assertEqual(type(self.decoder.readElement()), expected_type)

    def test_infinites(self):
        self.buf.truncate()
        self.buf.seek(0, 0)
        self.buf.append(b'\x00\xff\xf8\x00\x00\x00\x00\x00\x00')
        x = self.decoder.readElement()
        self.assertTrue(math.isnan(x))

        self.buf.truncate()
        self.buf.seek(0, 0)
        self.buf.append(b'\x00\xff\xf0\x00\x00\x00\x00\x00\x00')
        x = self.decoder.readElement()
        self.assertTrue(math.isinf(x) and x < 0)

        self.buf.truncate()
        self.buf.seek(0, 0)
        self.buf.append(b'\x00\x7f\xf0\x00\x00\x00\x00\x00\x00')
        x = self.decoder.readElement()
        self.assertTrue(math.isinf(x) and x > 0)

    def test_boolean(self):
        self.assertDecoded(True, b'\x01\x01')
        self.assertDecoded(False, b'\x01\x00')

    def test_string(self):
        self.assertDecoded('', b'\x02\x00\x00')
        self.assertDecoded('hello', b'\x02\x00\x05hello')
        self.assertDecoded(
            u'ᚠᛇᚻ',
            b'\x02\x00\t\xe1\x9a\xa0\xe1\x9b\x87\xe1\x9a\xbb'
        )

    def test_longstring(self):
        a = 'a' * 65537

        self.assertDecoded(a, b'\x0c\x00\x01\x00\x01' + a.encode('utf-8'))

    def test_null(self):
        self.assertDecoded(None, b'\x05')

    def test_list(self):
        self.assertDecoded([], b'\x0a\x00\x00\x00\x00')
        self.assertDecoded(
            [1, 2, 3],
            b'\x0a\x00\x00\x00\x03\x00\x3f\xf0\x00\x00\x00\x00\x00\x00\x00\x40'
            b'\x00\x00\x00\x00\x00\x00\x00\x00\x40\x08\x00\x00\x00\x00\x00\x00'
        )

    def test_dict(self):
        bytes = b'\x08\x00\x00\x00\x00\x00\x01\x61\x02\x00\x01\x61\x00\x00\x09'

        self.assertDecoded({'a': 'a'}, bytes)

        self.buf.append(bytes)
        self.decoder.readElement()

    def test_mixed_array(self):
        bytes = (
            b'\x08\x00\x00\x00\x00\x00\x01a\x00?\xf0\x00\x00\x00\x00\x00\x00'
            b'\x00\x01c\x00@\x08\x00\x00\x00\x00\x00\x00\x00\x01b\x00@\x00\x00'
            b'\x00\x00\x00\x00\x00\x00\x00\t'
        )

        self.assertDecoded(miniamf.MixedArray(a=1, b=2, c=3), bytes)

        self.buf.append(bytes)
        self.decoder.readElement()

    def test_date(self):
        self.assertDecoded(
            datetime.datetime(2005, 3, 18, 1, 58, 31),
            b'\x0bBp+6!\x15\x80\x00\x00\x00'
        )
        self.assertDecoded(
            datetime.datetime(2009, 3, 8, 23, 30, 47, 770122),
            b'\x0bBq\xfe\x86\xca5\xa1\xf4\x00\x00'
        )

    def test_xml(self):
        e = b'<a><b>hello world</b></a>'
        ret = self.decode(b'\x0f\x00\x00\x00\x19' + e)

        self.assertEqual(xml.tostring(ret), e)

    def test_xml_references(self):
        self.buf.truncate()
        self.buf.seek(0, 0)
        self.buf.append(
            b'\x0f\x00\x00\x00\x19<a><b>hello world</b></a>\x07\x00\x00'
        )

        self.assertEqual(
            xml.tostring(xml.fromstring('<a><b>hello world</b></a>')),
            xml.tostring(self.decoder.readElement()))

        self.assertEqual(
            xml.tostring(xml.fromstring('<a><b>hello world</b></a>')),
            xml.tostring(self.decoder.readElement()))

    def test_object(self):
        bytes = b'\x03\x00\x01a\x02\x00\x01b\x00\x00\x09'

        self.assertDecoded({'a': 'b'}, bytes)

        self.buf.append(bytes)
        self.decoder.readElement()

    def test_registered_class(self):
        miniamf.register_class(Spam, alias='org.miniamf.spam')

        bytes = (
            b'\x10\x00\x10org.miniamf.spam\x00\x03baz\x02\x00\x05hello\x00\x00'
            b'\x09'
        )

        obj = self.decode(bytes)

        self.assertEqual(type(obj), Spam)

        self.assertTrue(hasattr(obj, 'baz'))
        self.assertEqual(obj.baz, 'hello')

    def test_complex_list(self):
        x = datetime.datetime(2007, 11, 3, 8, 7, 37, 437000)

        self.assertDecoded(
            [['test', 'test', 'test', 'test']],
            b'\x0A\x00\x00\x00\x01\x0A\x00\x00\x00\x04\x02\x00\x04\x74\x65\x73'
            b'\x74\x02\x00\x04\x74\x65\x73\x74\x02\x00\x04\x74\x65\x73\x74\x02'
            b'\x00\x04\x74\x65\x73\x74'
        )
        self.assertDecoded(
            [x],
            b'\x0a\x00\x00\x00\x01\x0b\x42\x71\x60\x48\xcf\xed\xd0\x00\x00\x00'
        )
        self.assertDecoded(
            [[{'a': u'spam', 'b': u'eggs'}, {'a': u'spam', 'b': u'eggs'}]],
            b'\n\x00\x00\x00\x01\n\x00\x00\x00\x02\x08\x00\x00\x00\x00\x00\x01'
            b'a\x02\x00\x04spam\x00\x01b\x02\x00\x04eggs\x00\x00\t\x07\x00\x02'
        )
        self.assertDecoded(
            [[1.0]],
            b'\x0A\x00\x00\x00\x01\x0A\x00\x00\x00\x01\x00\x3F\xF0\x00\x00\x00'
            b'\x00\x00\x00'
        )

    def test_amf3(self):
        self.buf.append(b'\x11\x04\x01')
        self.assertEqual(self.decoder.readElement(), 1)

    def test_dynamic(self):
        class Foo(miniamf.ASObject):
            pass

        x = Foo()

        x.foo = 'bar'

        alias = miniamf.register_class(Foo, 'x')
        alias.exclude_attrs = ['hello']

        self.assertDecoded(
            x,
            b'\x10\x00\x01x\x00\x03foo\x02\x00\x03bar\x00\x05hello'
            b'\x02\x00\x05world\x00\x00\t'
        )

    def test_classic_class(self):
        miniamf.register_class(ClassicSpam, 'spam.eggs')

        self.buf.append(
            b'\x10\x00\tspam.eggs\x00\x03foo\x02\x00\x03bar\x00\x00\t'
        )

        foo = self.decoder.readElement()

        self.assertEqual(foo.foo, 'bar')

    def test_not_strict(self):
        self.assertFalse(self.decoder.strict)

        # write a typed object to the stream
        self.buf.append(
            b'\x10\x00\tspam.eggs\x00\x03foo\x02\x00\x03bar\x00\x00\t'
        )

        self.assertFalse('spam.eggs' in miniamf.CLASS_CACHE)

        obj = self.decoder.readElement()

        self.assertTrue(isinstance(obj, miniamf.TypedObject))
        self.assertEqual(obj.alias, 'spam.eggs')
        self.assertEqual(obj, {'foo': 'bar'})

    def test_strict(self):
        self.decoder.strict = True

        self.assertTrue(self.decoder.strict)

        # write a typed object to the stream
        self.buf.append(
            b'\x10\x00\tspam.eggs\x00\x03foo\x02\x00\x03bar\x00\x00\t'
        )

        self.assertFalse('spam.eggs' in miniamf.CLASS_CACHE)

        self.assertRaises(miniamf.UnknownClassAlias, self.decoder.readElement)

    def test_slots(self):
        class Person(object):
            __slots__ = ('family_name', 'given_name')

        self.buf.append(
            b'\x03\x00\x0bfamily_name\x02\x00\x03Doe\x00\ngiven_name\x02\x00'
            b'\x04Jane\x00\x00\t'
        )

        foo = self.decoder.readElement()

        self.assertEqual(foo.family_name, 'Doe')
        self.assertEqual(foo.given_name, 'Jane')

    def test_slots_registered(self):
        class Person(object):
            __slots__ = ('family_name', 'given_name')

        miniamf.register_class(Person, 'spam.eggs.Person')

        self.buf.append(
            b'\x10\x00\x10spam.eggs.Person\x00\x0bfamily_name\x02\x00\x03Doe'
            b'\x00\ngiven_name\x02\x00\x04Jane\x00\x00\t'
        )

        foo = self.decoder.readElement()

        self.assertTrue(isinstance(foo, Person))
        self.assertEqual(foo.family_name, 'Doe')
        self.assertEqual(foo.given_name, 'Jane')

    def test_ioerror_buffer_position(self):
        """
        Test to ensure that if an IOError is raised by `readElement` that
        the original position of the stream is restored.
        """
        bytes = miniamf.encode(u'foo', [1, 2, 3],
                               encoding=miniamf.AMF0).getvalue()

        self.buf.append(bytes[:-1])

        self.decoder.readElement()
        self.assertEqual(self.buf.tell(), 6)

        self.assertRaises(IOError, self.decoder.readElement)
        self.assertEqual(self.buf.tell(), 6)

    def test_timezone(self):
        self.decoder.timezone_offset = datetime.timedelta(hours=-5)

        self.buf.append(b'\x0bBr>\xc6\xf5w\x80\x00\x00\x00')

        f = self.decoder.readElement()

        self.assertEqual(f, datetime.datetime(2009, 9, 24, 9, 23, 23))

    def test_unsupported(self):
        self.assertDecoded(None, b'\x0D')

    def test_bad_reference(self):
        self.assertRaises(miniamf.ReferenceError,
                          self.decode, b'\x07\x00\x03')

    def test_iterate(self):
        self.assertRaises(StopIteration, lambda: next(self.decoder))

        self.decoder.send(b'\x02\x00\x00')
        self.decoder.send(b'\x02\x00\x05hello')
        self.decoder.send(b'\x02\x00\t\xe1\x9a\xa0\xe1\x9b\x87\xe1\x9a\xbb')

        self.assertEqual(next(self.decoder), '')
        self.assertEqual(next(self.decoder), 'hello')
        self.assertEqual(next(self.decoder), u'\u16a0\u16c7\u16bb')

        self.assertRaises(StopIteration, lambda: next(self.decoder))

        self.assertIdentical(iter(self.decoder), self.decoder)

    def test_bad_type(self):
        self.assertRaises(miniamf.DecodeError, self.decode, b'\xff')

    def test_kwargs(self):
        """
        Python 2 only accepts byte strings in kwargs.
        """
        def f(**kwargs):
            self.assertEqual(kwargs, {"a": "a"})

        kwargs = self.decode(b'\x03\x00\x01a\x02\x00\x01a\x00\x00\t')

        f(**kwargs)

    def test_numerical_keys_mixed_array(self):
        """
        Numerical keys in L{miniamf.MixedArray} must not cause a KeyError on
        decode.

        @see: #843
        """
        x = miniamf.MixedArray({'10': u'foobar'})

        bytes = miniamf.encode(x, encoding=miniamf.AMF0)

        d = list(miniamf.decode(bytes, encoding=miniamf.AMF0))

        self.assertEqual(d, [{10: u'foobar'}])

    def test_post_process(self):
        """
        Ensure that postprocessing happens when data has been decoded.
        """
        self.executed = False

        post_procs = miniamf.POST_DECODE_PROCESSORS[:]

        def restore_post_procs():
            miniamf.POST_DECODE_PROCESSORS = post_procs

        self.addCleanup(restore_post_procs)
        miniamf.POST_DECODE_PROCESSORS = []

        def postprocess(payload, context):
            self.assertEqual(payload, u'foo')
            self.assertEqual(context, {})

            self.executed = True

            return payload

        miniamf.add_post_decode_processor(postprocess)

        # setup complete
        bytes = miniamf.encode(u'foo', encoding=miniamf.AMF0).getvalue()

        self.decoder.send(bytes)
        ret = next(self.decoder)

        self.assertTrue(self.executed)
        self.assertEqual(ret, u'foo')


class RecordSetTestCase(unittest.TestCase, EncoderMixIn, DecoderMixIn):
    """
    Tests for L{amf0.RecordSet}
    """

    amf_type = miniamf.AMF0
    blob = (
        b'\x10\x00\tRecordSet\x00\nserverInfo\x03', (
            b'\x00\x06cursor\x00?\xf0\x00\x00\x00\x00\x00\x00',
            b'\x00\x0bcolumnNames\n\x00\x00\x00\x03\x02\x00\x01a\x02\x00\x01b'
            b'\x02\x00\x01c',
            b'\x00\x0binitialData\n\x00\x00\x00\x03\n\x00\x00\x00\x03\x00?\xf0'
            b'\x00\x00\x00\x00\x00\x00\x00@\x00\x00\x00\x00\x00\x00\x00\x00'
            b'@\x08\x00\x00\x00\x00\x00\x00\n\x00\x00\x00\x03\x00@\x10\x00'
            b'\x00\x00\x00\x00\x00\x00@\x14\x00\x00\x00\x00\x00\x00\x00@\x18'
            b'\x00\x00\x00\x00\x00\x00\n\x00\x00\x00\x03\x00@\x1c\x00\x00'
            b'\x00\x00\x00\x00\x00@ \x00\x00\x00\x00\x00\x00\x00@"\x00\x00'
            b'\x00\x00\x00\x00',
            b'\x00\x07version\x00?\xf0\x00\x00\x00\x00\x00\x00',
            b'\x00\ntotalCount\x00@\x08\x00\x00\x00\x00\x00\x00'
        ),
        b'\x00\x00\t\x00\x00\t'
    )

    def setUp(self):
        unittest.TestCase.setUp(self)
        EncoderMixIn.setUp(self)
        DecoderMixIn.setUp(self)

    def test_create(self):
        x = amf0.RecordSet()

        self.assertEqual(x.columns, [])
        self.assertEqual(x.items, [])
        self.assertEqual(x.service, None)
        self.assertEqual(x.id, None)

        x = amf0.RecordSet(columns=['spam', 'eggs'], items=[[1, 2]])

        self.assertEqual(x.columns, ['spam', 'eggs'])
        self.assertEqual(x.items, [[1, 2]])
        self.assertEqual(x.service, None)
        self.assertEqual(x.id, None)

        x = amf0.RecordSet(service={}, id=54)

        self.assertEqual(x.columns, [])
        self.assertEqual(x.items, [])
        self.assertEqual(x.service, {})
        self.assertEqual(x.id, 54)

    def test_server_info(self):
        # empty recordset
        x = amf0.RecordSet()

        si = x.serverInfo

        self.assertTrue(isinstance(si, dict))
        self.assertEqual(si.cursor, 1)
        self.assertEqual(si.version, 1)
        self.assertEqual(si.columnNames, [])
        self.assertEqual(si.initialData, [])
        self.assertEqual(si.totalCount, 0)

        try:
            si.serviceName
        except AttributeError:
            pass

        try:
            si.id
        except AttributeError:
            pass

        # basic create
        x = amf0.RecordSet(columns=['a', 'b', 'c'], items=[
            [1, 2, 3], [4, 5, 6], [7, 8, 9]])

        si = x.serverInfo

        self.assertTrue(isinstance(si, dict))
        self.assertEqual(si.cursor, 1)
        self.assertEqual(si.version, 1)
        self.assertEqual(si.columnNames, ['a', 'b', 'c'])
        self.assertEqual(si.initialData, [[1, 2, 3], [4, 5, 6], [7, 8, 9]])
        self.assertEqual(si.totalCount, 3)

        try:
            si.serviceName
        except AttributeError:
            pass

        try:
            si.id
        except AttributeError:
            pass

        # with service & id
        service = {'name': 'baz'}

        x = amf0.RecordSet(
            columns=['spam'], items=[['eggs']], service=service, id='asdfasdf'
        )

        si = x.serverInfo

        self.assertTrue(isinstance(si, dict))
        self.assertEqual(si.cursor, 1)
        self.assertEqual(si.version, 1)
        self.assertEqual(si.columnNames, ['spam'])
        self.assertEqual(si.initialData, [['eggs']])
        self.assertEqual(si.totalCount, 1)
        self.assertEqual(si.serviceName, 'baz')
        self.assertEqual(si.id, 'asdfasdf')

    def test_encode(self):
        self.buf = self.encoder.stream

        x = amf0.RecordSet(columns=['a', 'b', 'c'], items=[
            [1, 2, 3], [4, 5, 6], [7, 8, 9]])

        self.assertEncoded(x, self.blob)

    def test_decode(self):
        self.buf = self.decoder.stream
        x = self.decode(self.blob)

        self.assertTrue(isinstance(x, amf0.RecordSet))
        self.assertEqual(x.columns, ['a', 'b', 'c'])
        self.assertEqual(x.items, [[1, 2, 3], [4, 5, 6], [7, 8, 9]])
        self.assertEqual(x.service, None)
        self.assertEqual(x.id, None)

    def test_repr(self):
        x = amf0.RecordSet(
            columns=['spam'],
            items=[['eggs']],
            service={'name': 'baz'},
            id='asdfasdf',
        )

        self.assertEqual(
            repr(x),
            "<miniamf.amf0.RecordSet id=asdfasdf service={'name': 'baz'} at "
            "0x%x>" % (id(x),)
        )


class ClassInheritanceTestCase(ClassCacheClearingTestCase, EncoderMixIn):

    amf_type = miniamf.AMF0

    def setUp(self):
        # wtf
        ClassCacheClearingTestCase.setUp(self)
        EncoderMixIn.setUp(self)

    def test_simple(self):
        class A(object):
            class __amf__:
                static = ('a')

        class B(A):
            class __amf__:
                static = ('b')

        miniamf.register_class(A, 'A')
        miniamf.register_class(B, 'B')

        x = B()
        x.a = 'spam'
        x.b = 'eggs'

        self.assertEncoded(
            x,
            b'\x10\x00\x01B', (
                b'\x00\x01a\x02\x00\x04spam',
                b'\x00\x01b\x02\x00\x04eggs'
            ),
            b'\x00\x00\t'
        )

    def test_deep(self):
        class A(object):
            class __amf__:
                static = ('a')

        class B(A):
            class __amf__:
                static = ('b')

        class C(B):
            class __amf__:
                static = ('c')

        miniamf.register_class(A, 'A')
        miniamf.register_class(B, 'B')
        miniamf.register_class(C, 'C')

        x = C()
        x.a = 'spam'
        x.b = 'eggs'
        x.c = 'foo'

        self.assertEncoded(
            x,
            b'\x10\x00\x01C', (
                b'\x00\x01a\x02\x00\x04spam',
                b'\x00\x01c\x02\x00\x03foo',
                b'\x00\x01b\x02\x00\x04eggs'
            ),
            b'\x00\x00\t'
        )


class ExceptionEncodingTestCase(ClassCacheClearingTestCase):
    """
    Tests for encoding exceptions.
    """

    def setUp(self):
        ClassCacheClearingTestCase.setUp(self)

        self.buffer = util.BufferedByteStream()
        self.encoder = amf0.Encoder(self.buffer)

    def test_exception(self):
        try:
            raise Exception('foo bar')
        except Exception as e:
            self.encoder.writeElement(e)

        self.assertEqual(
            self.buffer.getvalue(),
            b'\x03\x00\x07message\x02\x00\x07foo bar\x00\x04name\x02\x00\t'
            b'Exception\x00\x00\t'
        )

    def test_user_defined(self):
        class FooBar(Exception):
            pass

        try:
            raise FooBar('foo bar')
        except Exception as e:
            self.encoder.writeElement(e)

        self.assertEqual(
            self.buffer.getvalue(),
            b'\x03\x00\x07message\x02\x00\x07foo bar\x00\x04name\x02\x00'
            b'\x06FooBar\x00\x00\t'
        )

    def test_typed(self):
        class XYZ(Exception):
            pass

        miniamf.register_class(XYZ, 'foo.bar')

        try:
            raise XYZ('blarg')
        except Exception as e:
            self.encoder.writeElement(e)

        self.assertEqual(
            self.buffer.getvalue(),
            b'\x10\x00\x07foo.bar\x00\x07message\x02\x00\x05blarg\x00\x04name'
            b'\x02\x00\x03XYZ\x00\x00\t'
        )



from miniamf.adapters import register_adapter


def when_imported(mod):
    """
    This function is called immediately after mymodule has been
    imported.  It configures Mini-AMF to encode a list when an instance
    of mymodule.CustomClass is encountered.
    """
    import miniamf

    miniamf.add_type(mod.CustomClass, lambda obj: list(obj))


register_adapter('mymodule', when_imported)



# Copyright (c) The PyAMF Project.
# See LICENSE.txt for details.

"""
Tests for the adapters module.

@since: 0.3.1
"""

import os
import sys

from miniamf import adapters
from miniamf.tests import util
from miniamf.tests.test_imports import ImportsTestCase


class AdapterHelperTestCase(ImportsTestCase):
    def setUp(self):
        ImportsTestCase.setUp(self)

        self.old_env = os.environ.copy()
        self.mods = sys.modules.copy()

        self.path = os.path.join(os.path.dirname(__file__), 'imports')
        sys.path.append(self.path)

    def tearDown(self):
        ImportsTestCase.tearDown(self)

        util.replace_dict(os.environ, self.old_env)
        util.replace_dict(sys.modules, self.mods)
        sys.path.remove(self.path)

    def test_basic(self):
        class Foo(object):
            def __call__(self, *args, **kwargs):
                pass

        def bar(*args, **kargs):
            pass

        self.assertRaises(TypeError, adapters.register_adapter, 'foo', 1)
        self.assertRaises(TypeError, adapters.register_adapter, 'foo', 'asdf')
        adapters.register_adapter('foo', Foo())
        adapters.register_adapter('foo', bar)
        adapters.register_adapter('foo', lambda x: x)

    def test_import(self):
        self.imported = False

        def x(mod):
            self.imported = True
            self.spam = mod

        adapters.register_adapter('spam', x)

        import spam

        self.assertTrue(self.imported)
        self.assertEqual(self.spam, spam)

    def test_get_adapter(self):
        from miniamf.adapters import _decimal

        self.assertTrue(adapters.get_adapter('decimal') is _decimal)



# Copyright (c) The PyAMF Project.
# See LICENSE.txt for details.

"""
AMF3 implementation.

C{AMF3} is the default serialization for
U{ActionScript<http://en.wikipedia.org/wiki/ActionScript>} 3.0 and
provides various advantages over L{AMF0<miniamf.amf0>}, which is used
for ActionScript 1.0 and 2.0. It adds support for sending C{int} and
C{uint} objects as integers and supports data types that are available
only in ActionScript 3.0, such as L{ByteArray} and L{ArrayCollection}.

@see: U{Official AMF3 Specification in English
<http://opensource.adobe.com/wiki/download/attachments/1114283/amf3_spec_05_05_08.pdf>}
@see: U{Official AMF3 Specification in Japanese
<http://opensource.adobe.com/wiki/download/attachments/1114283/JP_amf3_spec_121207.pdf>}
@see: U{AMF3 documentation on OSFlash
<http://osflash.org/documentation/amf3>}

@since: 0.1

"""

import datetime
import zlib

import miniamf
from . import codec, util, xml


__all__ = [
    'ByteArray',
    'Context',
    'Encoder',
    'Decoder',
]


#: The undefined type is represented by the undefined type marker. No further
#: information is encoded for this value.
TYPE_UNDEFINED = b'\x00'
#: The null type is represented by the null type marker. No further
#: information is encoded for this value.
TYPE_NULL = b'\x01'
#: The false type is represented by the false type marker and is used to
#: encode a Boolean value of C{false}. No further information is encoded for
#: this value.
TYPE_BOOL_FALSE = b'\x02'
#: The true type is represented by the true type marker and is used to encode
#: a Boolean value of C{true}. No further information is encoded for this
#: value.
TYPE_BOOL_TRUE = b'\x03'
#: In AMF 3 integers are serialized using a variable length signed 29-bit
#: integer.
#: @see: U{Parsing Integers on OSFlash (external)
#: <http://osflash.org/amf3/parsing_integers>}
TYPE_INTEGER = b'\x04'
#: This type is used to encode an ActionScript Number or an ActionScript
#: C{int} of value greater than or equal to 2^28 or an ActionScript uint of
#: value greater than or equal to 2^29. The encoded value is is always an 8
#: byte IEEE-754 double precision floating point value in network byte order
#: (sign bit in low memory). The AMF 3 number type is encoded in the same
#: manner as the AMF 0 L{Number<miniamf.amf0.TYPE_NUMBER>} type.
TYPE_NUMBER = b'\x05'
#: ActionScript String values are represented using a single string type in
#: AMF 3 - the concept of string and long string types from AMF 0 is not used.
#: Strings can be sent as a reference to a previously occurring String by
#: using an index to the implicit string reference table. Strings are encoding
#: using UTF-8 - however the header may either describe a string literal or a
#: string reference.
TYPE_STRING = b'\x06'
#: ActionScript 3.0 introduced a new XML type however the legacy C{XMLDocument}
#: type from ActionScript 1.0 and 2.0.is retained in the language as
#: C{flash.xml.XMLDocument}. Similar to AMF 0, the structure of an
#: C{XMLDocument} needs to be flattened into a string representation for
#: serialization. As with other strings in AMF, the content is encoded in
#: UTF-8. XMLDocuments can be sent as a reference to a previously occurring
#: C{XMLDocument} instance by using an index to the implicit object reference
#: table.
#: @see: U{OSFlash documentation (external)
#: <http://osflash.org/documentation/amf3
#:     #x07_-_xml_legacy_flash.xml.xmldocument_class>}
TYPE_XML = b'\x07'
#: In AMF 3 an ActionScript Date is serialized as the number of
#: milliseconds elapsed since the epoch of midnight, 1st Jan 1970 in the
#: UTC time zone. Local time zone information is not sent.
TYPE_DATE = b'\x08'
#: ActionScript Arrays are described based on the nature of their indices,
#: i.e. their type and how they are positioned in the Array.
TYPE_ARRAY = b'\x09'
#: A single AMF 3 type handles ActionScript Objects and custom user classes.
TYPE_OBJECT = b'\x0A'
#: ActionScript 3.0 introduces a new top-level XML class that supports
#: U{E4X<http://en.wikipedia.org/wiki/E4X>} syntax.
#: For serialization purposes the XML type needs to be flattened into a
#: string representation. As with other strings in AMF, the content is
#: encoded using UTF-8.
TYPE_XMLSTRING = b'\x0B'
#: ActionScript 3.0 introduces the L{ByteArray} type to hold an Array
#: of bytes. AMF 3 serializes this type using a variable length encoding
#: 29-bit integer for the byte-length prefix followed by the raw bytes
#: of the L{ByteArray}.
#: @see: U{Parsing ByteArrays on OSFlash (external)
#: <http://osflash.org/documentation/amf3/parsing_byte_arrays>}
TYPE_BYTEARRAY = b'\x0C'

#: Reference bit.
REFERENCE_BIT = 0x01

#: The maximum value for an int that will avoid promotion to an
#: ActionScript Number when sent via AMF 3 is represented by a
#: signed 29 bit integer: 2^28 - 1.
MAX_29B_INT = 0x0FFFFFFF

#: The minimum that can be represented by a signed 29 bit integer.
MIN_29B_INT = -0x10000000

ENCODED_INT_CACHE = {}


class ObjectEncoding:
    """
    AMF object encodings.
    """
    #: Property list encoding.
    #: The remaining integer-data represents the number of class members that
    #: exist. The property names are read as string-data. The values are then
    #: read as AMF3-data.
    STATIC = 0x00

    #: Externalizable object.
    #: What follows is the value of the "inner" object, including type code.
    #: This value appears for objects that implement IExternalizable, such as
    #: L{ArrayCollection} and L{ObjectProxy}.
    EXTERNAL = 0x01

    #: Name-value encoding.
    #: The property names and values are encoded as string-data followed by
    #: AMF3-data until there is an empty string property name. If there is a
    #: class-def reference there are no property names and the number of values
    #: is equal to the number of properties in the class-def.
    DYNAMIC = 0x02


class DataOutput(object):
    """
    I am a C{StringIO} type object containing byte data from the AMF stream.
    ActionScript 3.0 introduced the C{flash.utils.ByteArray} class to support
    the manipulation of raw data in the form of an Array of bytes.
    I provide a set of methods for writing binary data with ActionScript 3.0.

    This class is the I/O counterpart to the L{DataInput} class, which reads
    binary data.

    @see: U{IDataOutput on Adobe Help (external)
    <http://help.adobe.com/en_US/FlashPlatform/reference/actionscript/3/flash/utils/IDataOutput.html>}
    """

    def __init__(self, encoder):
        """
        @param encoder: Encoder containing the stream.
        @type encoder: L{amf3.Encoder<miniamf.amf3.Encoder>}
        """
        self.encoder = encoder
        self.stream = encoder.stream

    def writeBoolean(self, value):
        """
        Writes a Boolean value.

        @type value: C{bool}
        @param value: A C{Boolean} value determining which byte is written.
        If the parameter is C{True}, C{1} is written; if C{False}, C{0} is
        written.

        @raise ValueError: Non-boolean value found.
        """
        if not isinstance(value, bool):
            raise ValueError("Non-boolean value found")

        if value is True:
            self.stream.write_uchar(1)
        else:
            self.stream.write_uchar(0)

    def writeByte(self, value):
        """
        Writes a byte.

        @type value: C{int}
        """
        self.stream.write_char(value)

    def writeUnsignedByte(self, value):
        """
        Writes an unsigned byte.

        @type value: C{int}
        @since: 0.5
        """
        return self.stream.write_uchar(value)

    def writeDouble(self, value):
        """
        Writes an IEEE 754 double-precision (64-bit) floating
        point number.

        @type value: C{number}
        """
        self.stream.write_double(value)

    def writeFloat(self, value):
        """
        Writes an IEEE 754 single-precision (32-bit) floating
        point number.

        @type value: C{float}
        """
        self.stream.write_float(value)

    def writeInt(self, value):
        """
        Writes a 32-bit signed integer.

        @type value: C{int}
        """
        self.stream.write_long(value)

    def writeMultiByte(self, value, charset):
        """
        Writes a multibyte string to the datastream using the
        specified character set.

        @type value: string
        @param value: The string value to be written.
        @type charset: string
        @param charset: The string denoting the character set to use. Possible
            character set strings include C{shift-jis}, C{cn-gb},
            C{iso-8859-1} and others.
        @see: U{Supported character sets on Adobe Help (external)
            <http://help.adobe.com/en_US/FlashPlatform/reference/actionscript/3/charset-codes.html>}
        """
        if isinstance(value, str):
            value = value.encode(charset)

        self.stream.write(value)

    def writeObject(self, value):
        """
        Writes an object to data stream in AMF serialized format.

        @param value: The object to be serialized.
        """
        self.encoder.writeElement(value)

    def writeShort(self, value):
        """
        Writes a 16-bit integer.

        @type value: C{int}
        @param value: A byte value as an integer.
        """
        self.stream.write_short(value)

    def writeUnsignedShort(self, value):
        """
        Writes a 16-bit unsigned integer.

        @type value: C{int}
        @param value: A byte value as an integer.
        @since: 0.5
        """
        self.stream.write_ushort(value)

    def writeUnsignedInt(self, value):
        """
        Writes a 32-bit unsigned integer.

        @type value: C{int}
        @param value: A byte value as an unsigned integer.
        """
        self.stream.write_ulong(value)

    def writeUTF(self, value):
        """
        Writes a UTF-8 string to the data stream.

        The length of the UTF-8 string in bytes is written first,
        as a 16-bit integer, followed by the bytes representing the
        characters of the string.

        @type value: string
        @param value: The string value to be written.
        """
        if not isinstance(value, bytes):
            value = value.encode("utf-8")

        self.stream.write_ushort(len(value))
        self.stream.write(value)

    def writeUTFBytes(self, value):
        """
        Writes a UTF-8 string. Similar to L{writeUTF}, but does
        not prefix the string with a 16-bit length word.

        @type value: string
        @param value: The string value to be written.
        """
        if not isinstance(value, bytes):
            value = value.encode("utf-8")

        self.stream.write(value)


class DataInput(object):
    """
    I provide a set of methods for reading binary data with ActionScript 3.0.

    This class is the I/O counterpart to the L{DataOutput} class, which writes
    binary data.

    @see: U{IDataInput on Adobe Help (external)
    <http://help.adobe.com/en_US/FlashPlatform/reference/actionscript/3/flash/utils/IDataInput.html>}
    """

    def __init__(self, decoder=None):
        """
        @param decoder: AMF3 decoder containing the stream.
        @type decoder: L{amf3.Decoder<miniamf.amf3.Decoder>}
        """
        self.decoder = decoder
        self.stream = decoder.stream

    def readBoolean(self):
        """
        Read C{Boolean}.

        @raise ValueError: Error reading Boolean.
        @rtype: C{bool}
        @return: A Boolean value, C{True} if the byte
        is nonzero, C{False} otherwise.
        """
        byte = self.stream.read(1)

        if byte == b'\x00':
            return False
        if byte == b'\x01':
            return True
        raise ValueError("Error reading boolean")

    def readByte(self):
        """
        Reads a signed byte.

        @rtype: C{int}
        @return: The returned value is in the range -128 to 127.
        """
        return self.stream.read_char()

    def readDouble(self):
        """
        Reads an IEEE 754 double-precision floating point number from the
        data stream.

        @rtype: C{number}
        @return: An IEEE 754 double-precision floating point number.
        """
        return self.stream.read_double()

    def readFloat(self):
        """
        Reads an IEEE 754 single-precision floating point number from the
        data stream.

        @rtype: C{number}
        @return: An IEEE 754 single-precision floating point number.
        """
        return self.stream.read_float()

    def readInt(self):
        """
        Reads a signed 32-bit integer from the data stream.

        @rtype: C{int}
        @return: The returned value is in the range -2147483648 to 2147483647.
        """
        return self.stream.read_long()

    def readMultiByte(self, length, charset):
        """
        Reads a multibyte string of specified length from the data stream
        using the specified character set.

        @type length: C{int}
        @param length: The number of bytes from the data stream to read.
        @type charset: string
        @param charset: The string denoting the character set to use.

        @rtype: string
        @return: Unicode string
        """

        bytes = self.stream.read(length)
        return bytes.decode(charset)

    def readObject(self):
        """
        Reads an object from the data stream.

        @return: The deserialized object.
        """
        return self.decoder.readElement()

    def readShort(self):
        """
        Reads a signed 16-bit integer from the data stream.

        @rtype: C{uint}
        @return: The returned value is in the range -32768 to 32767.
        """
        return self.stream.read_short()

    def readUnsignedByte(self):
        """
        Reads an unsigned byte from the data stream.

        @rtype: C{uint}
        @return: The returned value is in the range 0 to 255.
        """
        return self.stream.read_uchar()

    def readUnsignedInt(self):
        """
        Reads an unsigned 32-bit integer from the data stream.

        @rtype: C{uint}
        @return: The returned value is in the range 0 to 4294967295.
        """
        return self.stream.read_ulong()

    def readUnsignedShort(self):
        """
        Reads an unsigned 16-bit integer from the data stream.

        @rtype: C{uint}
        @return: The returned value is in the range 0 to 65535.
        """
        return self.stream.read_ushort()

    def readUTF(self):
        """
        Reads a UTF-8 string from the data stream.

        The string is assumed to be prefixed with an unsigned
        short indicating the length in bytes.

        @rtype: Unicode string
        """
        length = self.stream.read_ushort()
        return self.stream.read_utf8_string(length)

    def readUTFBytes(self, length):
        """
        Reads a sequence of C{length} UTF-8 bytes from the data
        stream and returns a string.

        @type length: C{int}
        @param length: The number of bytes from the data stream to read.
        @rtype: Unicode string
        """
        return self.readMultiByte(length, "utf-8")


class ByteArray(util.BufferedByteStream, DataInput, DataOutput):
    """
    I am a C{StringIO} type object containing byte data from the AMF stream.
    ActionScript 3.0 introduced the C{flash.utils.ByteArray} class to support
    the manipulation of raw data in the form of an Array of bytes.

    Supports C{zlib} compression.

    Possible uses of the C{ByteArray} class:
     - Creating a custom protocol to connect to a client.
     - Writing your own AMF/Remoting packet.
     - Optimizing the size of your data by using custom data types.

    @see: U{ByteArray on Adobe Help (external)
    <http://help.adobe.com/en_US/FlashPlatform/reference/actionscript/3/flash/utils/ByteArray.html>}
    """

    _zlib_header = b'\x78\x9c'

    class __amf__:
        amf3 = True

    def __init__(self, buf=None):
        self.context = Context()

        util.BufferedByteStream.__init__(self, buf)
        DataInput.__init__(self, Decoder(self, self.context))
        DataOutput.__init__(self, Encoder(self, self.context))

        self.compressed = self.peek(2) == ByteArray._zlib_header

    def readObject(self):
        self.context.clear()

        return super(ByteArray, self).readObject()

    def writeObject(self, obj):
        self.context.clear()

        return super(ByteArray, self).writeObject(obj)

    def __eq__(self, other):
        if isinstance(other, ByteArray):
            return self.getvalue() == other.getvalue()
        return self.getvalue() == other

    def __ne__(self, other):
        if isinstance(other, ByteArray):
            return self.getvalue() != other.getvalue()
        return self.getvalue() != other

    def __lt__(self, other):
        if isinstance(other, ByteArray):
            return self.getvalue() < other.getvalue()
        return self.getvalue() < other

    def __le__(self, other):
        if isinstance(other, ByteArray):
            return self.getvalue() <= other.getvalue()
        return self.getvalue() <= other

    def __gt__(self, other):
        if isinstance(other, ByteArray):
            return self.getvalue() > other.getvalue()
        return self.getvalue() > other

    def __ge__(self, other):
        if isinstance(other, ByteArray):
            return self.getvalue() >= other.getvalue()
        return self.getvalue() >= other

    def encode(self):
        buf = self.getvalue()

        if not self.compressed:
            return buf

        buf = zlib.compress(buf)
        # FIXME nick: hacked
        return buf[0] + b'\xda' + buf[2:]

    def compress(self):
        """
        Forces compression of the underlying stream.
        """
        self.compressed = True


class ClassDefinition(object):
    """
    This is an internal class used by L{Encoder}/L{Decoder} to hold details
    about transient class trait definitions.
    """

    def __init__(self, alias):
        self.alias = alias
        self.reference = None

        alias.compile()

        self.attr_len = 0

        if alias.static_attrs:
            self.attr_len = len(alias.static_attrs)

        self.encoding = ObjectEncoding.DYNAMIC

        if alias.external:
            self.encoding = ObjectEncoding.EXTERNAL
        elif not alias.dynamic:
            if alias.encodable_properties is not None:
                if len(alias.static_attrs) == len(alias.encodable_properties):
                    self.encoding = ObjectEncoding.STATIC
            else:
                self.encoding = ObjectEncoding.STATIC

    def __repr__(self):
        my_repr = (
            '<%s.ClassDefinition reference=%r encoding=%r alias=%r at 0x%x>'
        )

        return my_repr % (
            self.__class__.__module__,
            self.reference,
            self.encoding,
            self.alias,
            id(self)
        )


class Context(codec.Context):
    """
    I hold the AMF3 context for en/decoding streams.

    @ivar strings: A list of string references.
    @type strings: L{codec.ByteStringReferenceCollection}
    @ivar classes: A list of L{ClassDefinition}.
    @type classes: C{list}
    """

    def __init__(self, **kwargs):
        self.strings = codec.ByteStringReferenceCollection()
        self.classes = {}
        self.class_ref = {}

        self.class_idx = 0

        codec.Context.__init__(self, **kwargs)

    def clear(self):
        """
        Clears the context.
        """
        codec.Context.clear(self)

        self.strings.clear()
        self.classes = {}
        self.class_ref = {}

        self.class_idx = 0

    def getString(self, ref):
        """
        Gets a string based on a reference C{ref}.

        @param ref: The reference index.
        @type ref: string

        @rtype: string or C{None}
        @return: The referenced string.
        """
        return self.strings.getByReference(ref)

    def getStringReference(self, s):
        """
        Return string reference.

        @type s: string
        @param s: The referenced string.
        @return: The reference index to the string.
        @rtype: C{int} or C{None}
        """
        return self.strings.getReferenceTo(s)

    def addString(self, s):
        """
        Creates a reference to C{s}. If the reference already exists, that
        reference is returned.

        @type s: byte string
        @param s: The string to be referenced.
        @rtype: C{int}
        @return: The reference index.

        @raise TypeError: The parameter C{s} is not a byte string.
        """
        if not isinstance(s, bytes):
            raise TypeError

        if len(s) == 0:
            return -1

        return self.strings.append(s)

    def getClassByReference(self, ref):
        """
        Return class reference.

        @return: Class reference.
        """
        return self.class_ref.get(ref)

    def getClass(self, klass):
        """
        Returns a class reference.

        @return: Class reference.
        """
        return self.classes.get(klass)

    def addClass(self, alias, klass):
        """
        Creates a reference to C{class_def}.

        @param alias: C{ClassDefinition} instance.
        @type alias: C{ClassDefinition}
        """
        ref = self.class_idx

        self.class_ref[ref] = alias
        cd = self.classes[klass] = alias

        cd.reference = ref

        self.class_idx += 1

        return ref


class Decoder(codec.Decoder):
    """
    Decodes an AMF3 data stream.
    """

    def buildContext(self, **kwargs):
        return Context(**kwargs)

    def getTypeFunc(self, data):
        if data == TYPE_UNDEFINED:
            return self.readUndefined
        if data == TYPE_NULL:
            return self.readNull
        if data == TYPE_BOOL_FALSE:
            return self.readBoolFalse
        if data == TYPE_BOOL_TRUE:
            return self.readBoolTrue
        if data == TYPE_INTEGER:
            return self.readInteger
        if data == TYPE_NUMBER:
            return self.readNumber
        if data == TYPE_STRING:
            return self.readString
        if data == TYPE_XML:
            return self.readXML
        if data == TYPE_DATE:
            return self.readDate
        if data == TYPE_ARRAY:
            return self.readArray
        if data == TYPE_OBJECT:
            return self.readObject
        if data == TYPE_XMLSTRING:
            return self.readXMLString
        if data == TYPE_BYTEARRAY:
            return self.readByteArray

    def readUndefined(self):
        """
        Read undefined.
        """
        return miniamf.Undefined

    def readNull(self):
        """
        Read null.

        @return: C{None}
        @rtype: C{None}
        """
        return None

    def readBoolFalse(self):
        """
        Returns C{False}.

        @return: C{False}
        @rtype: C{bool}
        """
        return False

    def readBoolTrue(self):
        """
        Returns C{True}.

        @return: C{True}
        @rtype: C{bool}
        """
        return True

    def readNumber(self):
        """
        Read number.
        """
        return self.stream.read_double()

    def readInteger(self, signed=True):
        """
        Reads and returns an integer from the stream.

        @type signed: C{bool}
        @see: U{Parsing integers on OSFlash
        <http://osflash.org/amf3/parsing_integers>} for the AMF3 integer data
        format.
        """
        return decode_int(self.stream, signed)

    def _readLength(self):
        x = decode_int(self.stream, False)

        return (x >> 1, x & REFERENCE_BIT == 0)

    def readBytes(self):
        """
        Reads and returns a utf-8 encoded byte array.
        """
        length, is_reference = self._readLength()

        if is_reference:
            return self.context.getString(length)

        if length == 0:
            return b''

        result = self.stream.read(length)
        self.context.addString(result)

        return result

    def readString(self):
        """
        Reads and returns a string from the stream.
        """
        length, is_reference = self._readLength()

        if is_reference:
            result = self.context.getString(length)

            return self.context.getStringForBytes(result)

        if length == 0:
            return ''

        result = self.stream.read(length)
        self.context.addString(result)

        return self.context.getStringForBytes(result)

    def readDate(self):
        """
        Read date from the stream.

        The timezone is ignored as the date is always in UTC.
        """
        ref = self.readInteger(False)

        if ref & REFERENCE_BIT == 0:
            return self.context.getObject(ref >> 1)

        ms = self.stream.read_double()
        result = util.get_datetime(ms / 1000.0)

        if self.timezone_offset is not None:
            result += self.timezone_offset

        self.context.addObject(result)

        return result

    def readArray(self):
        """
        Reads an array from the stream.

        @warning: There is a very specific problem with AMF3 where the first
        three bytes of an encoded empty C{dict} will mirror that of an encoded
        C{{'': 1, '2': 2}}
        """
        size = self.readInteger(False)

        if size & REFERENCE_BIT == 0:
            return self.context.getObject(size >> 1)

        size >>= 1

        key = self.readString()

        if key == '':
            # integer indexes only -> python list
            result = [self.readElement() for i in range(size)]

            self.context.addObject(result)

            return result

        result = miniamf.MixedArray()
        self.context.addObject(result)

        while key:
            result[key] = self.readElement()
            key = self.readString()

        for i in range(size):
            result[i] = self.readElement()

        return result

    def _getClassDefinition(self, ref):
        """
        Reads class definition from the stream.
        """
        is_ref = ref & REFERENCE_BIT == 0
        ref >>= 1

        if is_ref:
            return self.context.getClassByReference(ref)

        name = self.readString()
        alias = None

        if name == '':
            name = miniamf.ASObject

        try:
            alias = miniamf.get_class_alias(name)
        except miniamf.UnknownClassAlias:
            if self.strict:
                raise

            alias = miniamf.TypedObjectClassAlias(name)

        class_def = ClassDefinition(alias)

        class_def.encoding = ref & 0x03
        class_def.attr_len = ref >> 2
        class_def.static_properties = []

        if class_def.attr_len > 0:
            for i in range(class_def.attr_len):
                key = self.readString()
                class_def.static_properties.append(key)

        self.context.addClass(class_def, alias.klass)

        return class_def

    def _readStatic(self, class_def, obj):
        for attr in class_def.static_properties:
            obj[attr] = self.readElement()

    def _readDynamic(self, class_def, obj):
        attr = self.readString()

        while attr:
            obj[attr] = self.readElement()
            attr = self.readString()

    def readObject(self):
        """
        Reads an object from the stream.

        @raise ReferenceError: Unknown reference found.
        @raise DecodeError: Unknown object encoding detected.
        """
        ref = self.readInteger(False)

        if ref & REFERENCE_BIT == 0:
            obj = self.context.getObject(ref >> 1)

            if obj is None:
                raise miniamf.ReferenceError(
                    'Unknown reference %d' % (ref >> 1,)
                )

            return obj

        ref >>= 1

        class_def = self._getClassDefinition(ref)
        alias = class_def.alias

        obj = alias.createInstance(codec=self)
        obj_attrs = dict()

        self.context.addObject(obj)

        if class_def.encoding == ObjectEncoding.EXTERNAL:
            obj.__readamf__(DataInput(self))
            return obj
        elif class_def.encoding == ObjectEncoding.DYNAMIC:
            self._readStatic(class_def, obj_attrs)
            self._readDynamic(class_def, obj_attrs)
        elif class_def.encoding == ObjectEncoding.STATIC:
            self._readStatic(class_def, obj_attrs)
        else:
            raise miniamf.DecodeError("Unknown object encoding")

        alias.applyAttributes(obj, obj_attrs, codec=self)

        return obj

    def readXML(self):
        """
        Reads an xml object from the stream.

        @return: An etree interface compatible object
        @see: L{xml.set_default_interface}
        """
        ref = self.readInteger(False)

        if ref & REFERENCE_BIT == 0:
            return self.context.getObject(ref >> 1)

        x = xml.fromstring(
            self.stream.read(ref >> 1),
            forbid_dtd=self.context.forbid_dtd,
            forbid_entities=self.context.forbid_entities,
        )
        self.context.addObject(x)

        return x

    def readXMLString(self):
        """
        Reads a string from the data stream and converts it into
        an XML Tree.

        @see: L{readXML}
        """
        return self.readXML()

    def readByteArray(self):
        """
        Reads a string of data from the stream.

        Detects if the L{ByteArray} was compressed using C{zlib}.

        @see: L{ByteArray}
        @note: This is not supported in ActionScript 1.0 and 2.0.
        """
        ref = self.readInteger(False)

        if ref & REFERENCE_BIT == 0:
            return self.context.getObject(ref >> 1)

        buffer = self.stream.read(ref >> 1)
        compressed = False

        if buffer[0:2] == ByteArray._zlib_header:
            try:
                buffer = zlib.decompress(buffer)
                compressed = True
            except zlib.error:
                pass

        obj = ByteArray(buffer)
        obj.compressed = compressed

        self.context.addObject(obj)

        return obj


class Encoder(codec.Encoder):
    """
    Encodes an AMF3 data stream.
    """

    def __init__(self, *args, **kwargs):
        self.string_references = kwargs.pop('string_references', True)

        codec.Encoder.__init__(self, *args, **kwargs)

    def buildContext(self, **kwargs):
        return Context(**kwargs)

    def getTypeFunc(self, data):
        """
        @see: L{codec.Encoder.getTypeFunc}
        """
        t = type(data)

        if t is int:
            return self.writeInteger
        if t is ByteArray:
            return self.writeByteArray
        if t is miniamf.MixedArray:
            return self.writeDict

        return codec.Encoder.getTypeFunc(self, data)

    def writeUndefined(self, n):
        """
        Writes an C{miniamf.Undefined} value to the stream.
        """
        self.stream.write(TYPE_UNDEFINED)

    def writeNull(self, n):
        """
        Writes a C{null} value to the stream.
        """
        self.stream.write(TYPE_NULL)

    def writeBoolean(self, n):
        """
        Writes a Boolean to the stream.
        """
        t = TYPE_BOOL_FALSE if n is False else TYPE_BOOL_TRUE

        self.stream.write(t)

    def _writeInteger(self, n):
        """
        AMF3 integers are encoded.

        @param n: The integer data to be encoded to the AMF3 data stream.
        @type n: integer data

        @see: U{Parsing Integers on OSFlash
        <http://osflash.org/documentation/amf3/parsing_integers>}
        for more info.
        """
        self.stream.write(encode_int(n))

    def writeInteger(self, n):
        """
        Writes an integer to the stream.

        @type   n: integer data
        @param  n: The integer data to be encoded to the AMF3 data stream.
        """
        if n < MIN_29B_INT or n > MAX_29B_INT:
            self.writeNumber(float(n))

            return

        self.stream.write(TYPE_INTEGER)
        self.stream.write(encode_int(n))

    def writeNumber(self, n):
        """
        Writes a float to the stream.

        @type n: C{float}
        """
        self.stream.write(TYPE_NUMBER)
        self.stream.write_double(n)

    def serialiseBytes(self, b):
        if len(b) == 0:
            self.stream.write_uchar(REFERENCE_BIT)

            return

        if self.string_references:
            ref = self.context.getStringReference(b)

            if ref != -1:
                self._writeInteger(ref << 1)

                return

            self.context.addString(b)

        self._writeInteger((len(b) << 1) | REFERENCE_BIT)
        self.stream.write(b)

    def serialiseString(self, s):
        """
        Writes a raw string to the stream.

        @type   s: string
        @param  s: The string data to be encoded to the AMF3 data stream.
        """
        if not isinstance(s, bytes):
            s = self.context.getBytesForString(s)

        self.serialiseBytes(s)

    def writeBytes(self, b):
        """
        Writes a raw string to the stream.
        """
        self.stream.write(TYPE_STRING)

        self.serialiseBytes(b)

    def writeString(self, s):
        """
        Writes a string to the stream. It will be B{UTF-8} encoded.
        """
        s = self.context.getBytesForString(s)

        self.writeBytes(s)

    def writeDate(self, n):
        """
        Writes a C{datetime} instance to the stream.

        Does not support C{datetime.time} instances because AMF3 has
        no way to encode time objects, so please use C{datetime.datetime}
        instead.

        @type n: L{datetime}
        @param n: The C{Date} data to be encoded to the AMF3 data stream.
        @raise EncodeError: A datetime.time instance was found
        """
        if isinstance(n, datetime.time):
            raise miniamf.EncodeError(
                'A datetime.time instance was found but AMF3 has no way to '
                'encode time objects. Please use datetime.datetime instead '
                '(got:%r)' % (n,)
            )

        self.stream.write(TYPE_DATE)

        ref = self.context.getObjectReference(n)

        if ref != -1:
            self._writeInteger(ref << 1)

            return

        self.context.addObject(n)

        self.stream.write_uchar(REFERENCE_BIT)

        if self.timezone_offset is not None:
            n -= self.timezone_offset

        ms = util.get_timestamp(n)
        self.stream.write_double(ms * 1000.0)

    def writeList(self, n):
        """
        Writes a C{tuple}, C{set} or C{list} to the stream.

        @type n: One of C{__builtin__.tuple}, C{__builtin__.set}
            or C{__builtin__.list}
        @param n: The C{list} data to be encoded to the AMF3 data stream.
        """

        self.stream.write(TYPE_ARRAY)

        ref = self.context.getObjectReference(n)

        if ref != -1:
            self._writeInteger(ref << 1)

            return

        self.context.addObject(n)

        self._writeInteger((len(n) << 1) | REFERENCE_BIT)
        self.stream.write(b'\x01')

        [self.writeElement(x) for x in n]

    def writeDict(self, n):
        """
        Writes a C{dict} to the stream.

        @type n: C{__builtin__.dict}
        @param n: The C{dict} data to be encoded to the AMF3 data stream.
        @raise ValueError: Non int/string key value found in the C{dict}
        @raise EncodeError: C{dict} contains empty string keys.
        """
        # Design bug in AMF3 that cannot read/write empty key strings
        # for more info
        if '' in n:
            raise miniamf.EncodeError("dicts cannot contain empty string keys")

        self.stream.write(TYPE_ARRAY)

        ref = self.context.getObjectReference(n)

        if ref != -1:
            self._writeInteger(ref << 1)

            return

        self.context.addObject(n)

        # The AMF3 spec demands that all string indices be listed first
        int_keys = []
        str_keys = []

        for x in n.keys():
            if isinstance(x, int):
                int_keys.append(x)
            elif isinstance(x, str):
                str_keys.append(x)
            else:
                raise ValueError("Non integer/string key value found in dict")

        # Make sure the integer keys are within range
        l = len(int_keys)

        for x in int_keys:
            if l < x <= 0:
                # treat as a string key
                str_keys.append(x)
                del int_keys[int_keys.index(x)]

        int_keys.sort()

        # If integer keys don't start at 0, they will be treated as strings
        if len(int_keys) > 0 and int_keys[0] != 0:
            for x in int_keys:
                str_keys.append(str(x))
                del int_keys[int_keys.index(x)]

        self._writeInteger(len(int_keys) << 1 | REFERENCE_BIT)

        str_keys.sort()

        for x in str_keys:
            self.serialiseString(x)
            self.writeElement(n[x])

        self.stream.write_uchar(0x01)

        for k in int_keys:
            self.writeElement(n[k])

    def writeObject(self, obj):
        """
        Writes an object to the stream.
        """

        self.stream.write(TYPE_OBJECT)

        ref = self.context.getObjectReference(obj)

        if ref != -1:
            self._writeInteger(ref << 1)

            return

        self.context.addObject(obj)

        # object is not referenced, serialise it
        kls = obj.__class__
        definition = self.context.getClass(kls)
        alias = None
        class_ref = False  # if the class definition is a reference

        if definition:
            class_ref = True
            alias = definition.alias
        else:
            alias = self.context.getClassAlias(kls)
            definition = ClassDefinition(alias)

            self.context.addClass(definition, alias.klass)

        if class_ref:
            self.stream.write(definition.reference)
        else:
            ref = 0

            if definition.encoding != ObjectEncoding.EXTERNAL:
                ref += definition.attr_len << 4

            final_reference = encode_int(
                ref |
                definition.encoding << 2 |
                REFERENCE_BIT << 1 |
                REFERENCE_BIT
            )

            self.stream.write(final_reference)

            definition.reference = encode_int(
                definition.reference << 2 | REFERENCE_BIT)

            if alias.anonymous:
                self.stream.write(b'\x01')
            else:
                self.serialiseString(alias.alias)

            # work out what the final reference for the class will be.
            # this is okay because the next time an object of the same
            # class is encoded, class_ref will be True and never get here
            # again.

        if alias.external:
            obj.__writeamf__(DataOutput(self))

            return

        attrs = alias.getEncodableAttributes(obj, codec=self)

        if alias.static_attrs:
            sattrs = sorted(alias.static_attrs)
            if not class_ref:
                for attr in sattrs:
                    self.serialiseString(attr)

            for attr in sattrs:
                value = attrs.pop(attr)

                self.writeElement(value)

            if definition.encoding == ObjectEncoding.STATIC:
                return

        if definition.encoding == ObjectEncoding.DYNAMIC:
            if attrs:
                try:
                    keys = sorted(attrs)
                except TypeError:
                    raise miniamf.EncodeError(
                        'Unable to encode %r (unsortable attrs: %r)'
                        % (obj, attrs))

                for key in keys:
                    value = attrs[key]
                    if isinstance(key, int):
                        key = str(key)
                    elif not isinstance(key, str):
                        raise miniamf.EncodeError(
                            'Unable to encode %r (key %r not supported)'
                            % (obj, key))

                    self.serialiseString(key)
                    self.writeElement(value)

            self.stream.write(b'\x01')

    def writeByteArray(self, n):
        """
        Writes a L{ByteArray} to the data stream.

        @param n: The L{ByteArray} data to be encoded to the AMF3 data stream.
        @type n: L{ByteArray}
        """
        self.stream.write(TYPE_BYTEARRAY)

        ref = self.context.getObjectReference(n)

        if ref != -1:
            self._writeInteger(ref << 1)

            return

        self.context.addObject(n)

        buf = n.encode()
        self._writeInteger(len(buf) << 1 | REFERENCE_BIT)
        self.stream.write(buf)

    def writeXML(self, n):
        """
        Writes a XML string to the data stream.

        @type   n: L{ET<xml.ET>}
        @param  n: The XML Document to be encoded to the AMF3 data stream.
        """
        self.stream.write(TYPE_XMLSTRING)
        ref = self.context.getObjectReference(n)

        if ref != -1:
            self._writeInteger(ref << 1)

            return

        self.context.addObject(n)

        self.serialiseString(xml.tostring(n))


def encode_int(n):
    """
    Encodes an int as a variable length signed 29-bit integer as defined by
    the spec.

    @param n: The integer to be encoded
    @return: The encoded string
    @raise OverflowError: C{c} is out of range.
    """
    try:
        return ENCODED_INT_CACHE[n]
    except KeyError:
        pass

    if n < MIN_29B_INT or n > MAX_29B_INT:
        raise OverflowError("Out of range")

    if n < 0:
        n += 0x20000000

    encoded = bytearray()
    real_value = None

    if n > 0x1fffff:
        real_value = n
        n >>= 1
        encoded.append(0x80 | ((n >> 21) & 0xff))

    if n > 0x3fff:
        encoded.append(0x80 | ((n >> 14) & 0xff))

    if n > 0x7f:
        encoded.append(0x80 | ((n >> 7) & 0xff))

    if real_value is not None:
        n = real_value

    if n > 0x1fffff:
        encoded.append(n & 0xff)
    else:
        encoded.append(n & 0x7f)

    encoded = bytes(encoded)
    ENCODED_INT_CACHE.__setitem__(n, encoded)
    return encoded


def decode_int(stream, signed=False):
    """
    Decode C{int}.
    """
    n = result = 0
    b = stream.read_uchar()

    while b & 0x80 != 0 and n < 3:
        result <<= 7
        result |= b & 0x7f
        b = stream.read_uchar()
        n += 1

    if n < 3:
        result <<= 7
        result |= b
    else:
        result <<= 8
        result |= b

        if result & 0x10000000 != 0:
            if signed:
                result -= 0x20000000
            else:
                result <<= 1
                result += 1

    return result


miniamf.register_class(ByteArray)



import pyamf

@RemoteClass(alias="model.MyClass")
class MyClass:
    def __init__(self, *args, **kwargs):
        self.a = args[0]
        self.b = args[1]

class RemoteClass(object):
    def __init__(self, alias):
        self.alias = alias

    def __call__(self, klass):
        pyamf.register_class(klass, self.alias)
        return klass



# Copyright (c) The PyAMF Project.
# See LICENSE.txt for details.

"""
Provides XML support.

@since: 0.6
"""

#: list of supported third party packages that support the C{etree}
#: interface. At least enough for our needs anyway.
from miniamf.util import get_module

ETREE_MODULES = ['defusedxml.ElementTree']

#: A tuple of class/type objects that are used to represent XML objects.
types = None
#: A mapping of type -> module for all known xml types.
modules = {}
#: The module that will be used to create C{ElementTree} instances.
ET = None

__all__ = ['set_default_interface']


def set_default_interface(etree):
    """
    Sets the default interface that Mini-AMF will use to deal with XML entities
    (both objects and blobs).
    """
    global types, ET, modules

    t = _get_etree_type(etree)

    _types = set(types or [])
    _types.update([t])

    types = tuple(_types)

    modules[t] = etree

    old, ET = ET, etree

    return old


def find_libs():
    """
    Run through L{ETREE_MODULES} and find C{ElementTree} implementations so
    that any type can be encoded.

    We work through the C implementations first, then the pure Python versions.
    The downside to this is that B{all} libraries will be imported but I{only}
    one is ever used. The libs are small (relatively) and the flexibility that
    this gives seems to outweigh the cost. Time will tell.
    """

    types = []
    mapping = {}

    for mod in ETREE_MODULES:
        try:
            etree = get_module(mod)
        except ImportError:
            continue

        t = _get_etree_type(etree)

        types.append(t)
        mapping[t] = etree

    return tuple(types), mapping


def is_xml(obj):
    """
    Determines C{obj} is a valid XML type.

    If L{types} is not populated then L{find_libs} be called.
    """
    try:
        _bootstrap()
    except ImportError:
        return False

    return isinstance(obj, types)


def _get_type(e):
    """
    Returns the type associated with handling XML objects from this etree
    interface.
    """
    try:
        return e.__class__
    except AttributeError:
        return type(e)


def _get_etree_type(etree):
    """
    Returns the type associated with handling XML objects from this etree
    interface.
    """
    e = etree.fromstring('<foo/>')

    return _get_type(e)


def _no_et():
    raise ImportError(
        'Unable to find at least one compatible ElementTree library, use '
        'miniamf.set_default_etree to enable XML support'
    )


def _bootstrap():
    global types, modules

    if types is None:
        types, modules = find_libs()

    if ET is None:
        try:
            etree = modules[types[0]]
        except IndexError:
            _no_et()

        set_default_interface(etree)


def tostring(element, *args, **kwargs):
    """
    Helper func to provide easy access to the (possibly) moving target that is
    C{ET}.
    """
    _bootstrap()
    t = _get_type(element)

    etree = modules.get(t, None)

    if not etree:
        raise RuntimeError(
            'Unable to find the etree implementation related to %r '
            '(type %r)' % (element, t)
        )

    return etree.tostring(element, *args, **kwargs)


def fromstring(*args, **kwargs):
    """
    Helper func to provide easy access to the (possibly) moving target that is
    C{ET}.
    """
    _bootstrap()

    kwargs.setdefault('forbid_dtd', True)
    kwargs.setdefault('forbid_entities', True)

    return ET.fromstring(*args, **kwargs)
